#!/usr/bin/env python

import argparse
import asyncio
import grp
import logging
import os
import pwd
import subprocess
from pathlib import Path
from typing import Coroutine, List

import path_util  # noqa: F401

from bin.hummingbot import UIStartListener, detect_available_port
from hummingbot import init_logging
from hummingbot.client.command.start_command import GATEWAY_READY_TIMEOUT
from hummingbot.client.config.config_crypt import BaseSecretsManager, ETHKeyFileSecretManger
from hummingbot.client.config.config_helpers import (
    ClientConfigAdapter,
    all_configs_complete,
    create_yml_files_legacy,
    load_client_config_map_from_file,
    load_strategy_config_map_from_file,
    read_system_configs_from_yml,
)
from hummingbot.client.config.security import Security
from hummingbot.client.hummingbot_application import HummingbotApplication
from hummingbot.client.settings import SCRIPT_STRATEGY_CONF_DIR_PATH, STRATEGIES_CONF_DIR_PATH, AllConnectorSettings
from hummingbot.client.ui import login_prompt
from hummingbot.client.ui.style import load_style
from hummingbot.core.event.events import HummingbotUIEvent
from hummingbot.core.management.console import start_management_console
from hummingbot.core.utils.async_utils import safe_gather


class CmdlineParser(argparse.ArgumentParser):
    def __init__(self):
        super().__init__()
        self.add_argument("--config-file-name", "-f",
                          type=str,
                          required=False,
                          help="Specify a file in `conf/` to load as the strategy config file.")
        self.add_argument("--v2",
                          type=str,
                          required=False,
                          dest="v2_conf",
                          help="V2 strategy config file name (from conf/scripts/).")
        self.add_argument("--config-password", "-p",
                          type=str,
                          required=False,
                          help="Specify the password to unlock your encrypted files.")
        self.add_argument("--auto-set-permissions",
                          type=str,
                          required=False,
                          help="Try to automatically set config / logs / data dir permissions, "
                               "useful for Docker containers.")
        self.add_argument("--headless",
                          type=bool,
                          nargs='?',
                          const=True,
                          default=None,
                          help="Run in headless mode without CLI interface.")


def autofix_permissions(user_group_spec: str):
    uid, gid = [sub_str for sub_str in user_group_spec.split(':')]

    uid = int(uid) if uid.isnumeric() else pwd.getpwnam(uid).pw_uid
    gid = int(gid) if gid.isnumeric() else grp.getgrnam(gid).gr_gid

    os.environ["HOME"] = pwd.getpwuid(uid).pw_dir
    project_home: str = os.path.realpath(os.path.join(__file__, "../../"))

    gateway_path: str = Path.home().joinpath(".hummingbot-gateway").as_posix()
    subprocess.run(
        f"cd '{project_home}' && "
        f"sudo chown -R {user_group_spec} conf/ data/ logs/ scripts/ {gateway_path}",
        capture_output=True,
        shell=True
    )
    os.setgid(gid)
    os.setuid(uid)


async def quick_start(args: argparse.Namespace, secrets_manager: BaseSecretsManager):
    """Start Hummingbot using unified HummingbotApplication in either UI or headless mode."""
    client_config_map = load_client_config_map_from_file()

    if args.auto_set_permissions is not None:
        autofix_permissions(args.auto_set_permissions)

    if not Security.login(secrets_manager):
        logging.getLogger().error("Invalid password.")
        return

    await Security.wait_til_decryption_done()
    await create_yml_files_legacy()
    # Initialize logging with basic setup first - will be re-initialized later with correct strategy file name if needed
    init_logging("hummingbot_logs.yml", client_config_map)
    await read_system_configs_from_yml()

    # Automatically enable MQTT autostart for headless mode
    if args.headless:
        client_config_map.mqtt_bridge.mqtt_autostart = True

    AllConnectorSettings.initialize_paper_trade_settings(client_config_map.paper_trade.paper_trade_exchanges)

    # Create unified application that handles both headless and UI modes
    hb = HummingbotApplication.main_application(client_config_map=client_config_map, headless_mode=args.headless)

    # Load and start strategy if provided
    if args.v2_conf is not None or args.config_file_name is not None:
        success = await load_and_start_strategy(hb, args)
        if not success:
            logging.getLogger().error("Failed to load strategy. Exiting.")
            raise SystemExit(1)

    await wait_for_gateway_ready(hb)

    # Run the application
    await run_application(hb, args, client_config_map)


async def wait_for_gateway_ready(hb):
    """Wait until the gateway is ready before starting the strategy."""
    exchange_settings = [
        AllConnectorSettings.get_connector_settings().get(e, None)
        for e in hb.trading_core.connector_manager.connectors.keys()
    ]
    uses_gateway = any([s.uses_gateway_generic_connector() for s in exchange_settings])
    if not uses_gateway:
        return
    try:
        await asyncio.wait_for(hb.trading_core.gateway_monitor.ready_event.wait(), timeout=GATEWAY_READY_TIMEOUT)
    except asyncio.TimeoutError:
        logging.getLogger().error(
            f"TimeoutError waiting for gateway service to go online... Please ensure Gateway is configured correctly."
            f"Unable to start strategy {hb.trading_core.strategy_name}. ")
        raise


async def load_and_start_strategy(hb: HummingbotApplication, args: argparse.Namespace,):
    """Load and start strategy based on file type and mode."""
    import yaml

    if args.v2_conf:
        # V2 config-driven start: derive script from config file
        conf_path = SCRIPT_STRATEGY_CONF_DIR_PATH / args.v2_conf
        if not conf_path.exists():
            logging.getLogger().error(f"V2 config file not found: {conf_path}")
            return False

        with open(conf_path) as f:
            config_data = yaml.safe_load(f) or {}
        script_file = config_data.get("script_file_name", "")
        if not script_file:
            logging.getLogger().error("Config file is missing 'script_file_name' field.")
            return False

        strategy_name = script_file.replace(".py", "")
        hb.strategy_file_name = args.v2_conf
        hb.trading_core.strategy_name = strategy_name

        if args.headless:
            logging.getLogger().info(f"Starting V2 script strategy: {strategy_name}")
            success = await hb.trading_core.start_strategy(
                strategy_name,
                args.v2_conf,
                args.v2_conf
            )
            if not success:
                logging.getLogger().error("Failed to start strategy")
                return False
        else:
            # UI mode - trigger start via listener
            hb.script_config = args.v2_conf

    elif args.config_file_name is not None:
        # Regular strategy with YAML config (V1 flow)
        hb.strategy_file_name = args.config_file_name.split(".")[0]  # Remove .yml extension

        try:
            strategy_config = await load_strategy_config_map_from_file(
                STRATEGIES_CONF_DIR_PATH / args.config_file_name
            )
        except FileNotFoundError:
            logging.getLogger().error(f"Strategy config file not found: {STRATEGIES_CONF_DIR_PATH / args.config_file_name}")
            return False
        except Exception as e:
            logging.getLogger().error(f"Error loading strategy config file: {e}")
            return False

        strategy_name = (
            strategy_config.strategy
            if isinstance(strategy_config, ClientConfigAdapter)
            else strategy_config.get("strategy").value
        )
        hb.trading_core.strategy_name = strategy_name

        if args.headless:
            logging.getLogger().info(f"Starting regular strategy: {strategy_name}")
            success = await hb.trading_core.start_strategy(
                strategy_name,
                strategy_config,
                args.config_file_name
            )
            if not success:
                logging.getLogger().error("Failed to start strategy")
                return False
        else:
            # UI mode - set properties for UIStartListener
            hb.strategy_config_map = strategy_config

            # Check if config is complete for UI mode
            if not all_configs_complete(strategy_config, hb.client_config_map):
                hb.status()

    return True


async def run_application(hb: HummingbotApplication, args: argparse.Namespace, client_config_map):
    """Run the application in headless or UI mode."""
    if args.headless:
        # Re-initialize logging with proper strategy file name for headless mode
        from hummingbot import init_logging
        log_file_name = hb.strategy_file_name.split(".")[0] if hb.strategy_file_name else "hummingbot"
        init_logging("hummingbot_logs.yml", hb.client_config_map,
                     override_log_level=hb.client_config_map.log_level,
                     strategy_file_path=log_file_name)
        await hb.run()
    else:
        # Set up UI mode with start listener
        start_listener: UIStartListener = UIStartListener(
            hb,
            is_script=args.v2_conf is not None,
            script_config=getattr(hb, 'script_config', None),
            is_quickstart=True
        )
        hb.app.add_listener(HummingbotUIEvent.Start, start_listener)

        tasks: List[Coroutine] = [hb.run()]
        if client_config_map.debug_console:
            management_port: int = detect_available_port(8211)
            tasks.append(start_management_console(locals(), host="localhost", port=management_port))

        await safe_gather(*tasks)


def main():
    args = CmdlineParser().parse_args()

    # Parse environment variables from Dockerfile.
    # If an environment variable is not empty and it's not defined in the arguments, then we'll use the environment
    # variable.
    if args.config_file_name is None and len(os.environ.get("CONFIG_FILE_NAME", "")) > 0:
        args.config_file_name = os.environ["CONFIG_FILE_NAME"]

    if args.v2_conf is None and len(os.environ.get("SCRIPT_CONFIG", "")) > 0:
        args.v2_conf = os.environ["SCRIPT_CONFIG"]

    if args.config_password is None and len(os.environ.get("CONFIG_PASSWORD", "")) > 0:
        args.config_password = os.environ["CONFIG_PASSWORD"]

    if args.headless is None and len(os.environ.get("HEADLESS_MODE", "")) > 0:
        args.headless = os.environ["HEADLESS_MODE"].lower() == "true"

    # If no password is given from the command line, prompt for one.
    secrets_manager_cls = ETHKeyFileSecretManger
    client_config_map = load_client_config_map_from_file()
    if args.config_password is None:
        secrets_manager = login_prompt(secrets_manager_cls, style=load_style(client_config_map))
        if not secrets_manager:
            return
    else:
        secrets_manager = secrets_manager_cls(args.config_password)

    try:
        ev_loop: asyncio.AbstractEventLoop = asyncio.get_running_loop()
    except RuntimeError:
        ev_loop: asyncio.AbstractEventLoop = asyncio.new_event_loop()
        asyncio.set_event_loop(ev_loop)

    ev_loop.run_until_complete(quick_start(args, secrets_manager))


if __name__ == "__main__":
    main()



from bidict import bidict

from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit

REST_URL = "https://api.kraken.com"
HEALTH_CHECK_ENDPOINT = "/0/public/Time"
CANDLES_ENDPOINT = "/0/public/OHLC"
WS_CANDLES_ENDPOINT = "ohlc"

WSS_URL = "wss://ws.kraken.com"
KRAKEN_TO_HB_MAP = {
    "XBT": "BTC",
    "XDG": "DOGE",
}

INTERVALS = bidict({
    "1m": "1",
    "5m": "5",
    "15m": "15",
    "30m": "30",
    "1h": "60",
    "4h": "240",
    "1d": "1440",
    "1w": "10080",
})
MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST = MAX_CANDLES_AGO = 720
PUBLIC_ENDPOINT_LIMIT_ID = "PublicPoints"
RATE_LIMITS = [
    RateLimit(
        limit_id=PUBLIC_ENDPOINT_LIMIT_ID,
        limit=1,
        time_interval=1,
    ),
    # Public Endpoints
    RateLimit(
        limit_id=CANDLES_ENDPOINT,
        limit=1,
        time_interval=1,
        linked_limits=[LinkedLimitWeightPair(PUBLIC_ENDPOINT_LIMIT_ID)],
    ),
    RateLimit(
        limit_id=HEALTH_CHECK_ENDPOINT,
        limit=1,
        time_interval=1,
        linked_limits=[LinkedLimitWeightPair(PUBLIC_ENDPOINT_LIMIT_ID)],
    )
]



import uuid
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from hummingbot.connector.exchange.binance import binance_constants as CONSTANTS
from hummingbot.connector.exchange.binance.binance_auth import BinanceAuth
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.web_assistant.connections.data_types import WSJSONRequest, WSResponse
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.exchange.binance.binance_exchange import BinanceExchange


class BinanceAPIUserStreamDataSource(UserStreamTrackerDataSource):

    _logger: Optional[HummingbotLogger] = None

    def __init__(self,
                 auth: BinanceAuth,
                 trading_pairs: List[str],
                 connector: 'BinanceExchange',
                 api_factory: WebAssistantsFactory,
                 domain: str = CONSTANTS.DEFAULT_DOMAIN):
        super().__init__()
        self._auth: BinanceAuth = auth
        self._domain = domain
        self._api_factory = api_factory
        self._connector = connector

    async def _get_ws_assistant(self) -> WSAssistant:
        return await self._api_factory.get_ws_assistant()

    async def _connected_websocket_assistant(self) -> WSAssistant:
        ws = await self._get_ws_assistant()
        url = CONSTANTS.WSS_API_URL.format(self._domain)
        await ws.connect(ws_url=url, ping_timeout=CONSTANTS.WS_HEARTBEAT_TIME_INTERVAL)
        return ws

    async def _subscribe_channels(self, websocket_assistant: WSAssistant):
        try:
            params = self._auth.generate_ws_subscribe_params()
            request_id = str(uuid.uuid4())
            payload = {
                "id": request_id,
                "method": "userDataStream.subscribe.signature",
                "params": params,
            }
            subscribe_request = WSJSONRequest(payload=payload)
            await websocket_assistant.send(subscribe_request)

            response: WSResponse = await websocket_assistant.receive()
            data = response.data

            if not isinstance(data, dict) or data.get("status") != 200:
                raise IOError(f"Error subscribing to user stream (response: {data})")

            self.logger().info("Successfully subscribed to user data stream via WebSocket API")
        except IOError:
            raise
        except Exception:
            self.logger().exception("Unexpected error subscribing to user data stream")
            raise

    async def _process_event_message(self, event_message: Dict[str, Any], queue):
        if not isinstance(event_message, dict) or len(event_message) == 0:
            return
        # Filter out WebSocket API response messages (subscribe confirmations, etc.)
        if "id" in event_message and "status" in event_message:
            return
        # Unwrap WS API event container: {"subscriptionId": N, "event": {...}}
        if "event" in event_message and "subscriptionId" in event_message:
            event_message = event_message["event"]
        # Handle stream termination by triggering reconnect
        if event_message.get("e") == "eventStreamTerminated":
            raise ConnectionError("User data stream subscription terminated by server")
        queue.put_nowait(event_message)

    async def _on_user_stream_interruption(self, websocket_assistant: Optional[WSAssistant]):
        websocket_assistant and await websocket_assistant.disconnect()



from decimal import Decimal

from pydantic import BaseModel, ConfigDict, field_validator

from hummingbot.core.data_type.common import TradeType
from hummingbot.strategy_v2.utils.common import parse_enum_value


class InitialPositionConfig(BaseModel):
    """
    Configuration for an initial position that the controller should consider.
    This is used when the user already has assets in their account and wants
    the controller to manage them.
    """
    connector_name: str
    trading_pair: str
    amount: Decimal
    side: TradeType

    @field_validator('side', mode='before')
    @classmethod
    def parse_side(cls, v):
        """Parse side field from string to TradeType enum."""
        return parse_enum_value(TradeType, v, "side")

    model_config = ConfigDict(arbitrary_types_allowed=True)



import asyncio
import logging
from decimal import Decimal
from typing import Dict, List, Optional, Union

from hummingbot.connector.connector_base import ConnectorBase
from hummingbot.core.data_type.common import OrderType, PositionAction, PriceType, TradeType
from hummingbot.core.data_type.order_candidate import OrderCandidate, PerpetualOrderCandidate
from hummingbot.core.event.events import (
    BuyOrderCompletedEvent,
    BuyOrderCreatedEvent,
    MarketOrderFailureEvent,
    OrderCancelledEvent,
    OrderFilledEvent,
    SellOrderCompletedEvent,
    SellOrderCreatedEvent,
)
from hummingbot.logger import HummingbotLogger
from hummingbot.strategy.strategy_v2_base import StrategyV2Base
from hummingbot.strategy_v2.executors.executor_base import ExecutorBase
from hummingbot.strategy_v2.executors.position_executor.data_types import PositionExecutorConfig
from hummingbot.strategy_v2.models.base import RunnableStatus
from hummingbot.strategy_v2.models.executors import CloseType, TrackedOrder


class PositionExecutor(ExecutorBase):
    _logger = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(__name__)
        return cls._logger

    def __init__(self, strategy: StrategyV2Base, config: PositionExecutorConfig,
                 update_interval: float = 1.0, max_retries: int = 10):
        """
        Initialize the PositionExecutor instance.

        :param strategy: The strategy to be used by the PositionExecutor.
        :param config: The configuration for the PositionExecutor, subclass of PositionExecutoConfig.
        :param update_interval: The interval at which the PositionExecutor should be updated, defaults to 1.0.
        :param max_retries: The maximum number of retries for the PositionExecutor, defaults to 5.
        """
        if config.triple_barrier_config.time_limit_order_type != OrderType.MARKET or \
                config.triple_barrier_config.stop_loss_order_type != OrderType.MARKET:
            error = "Only market orders are supported for time_limit and stop_loss"
            self.logger().error(error)
            raise ValueError(error)
        super().__init__(strategy=strategy, config=config, connectors=[config.connector_name],
                         update_interval=update_interval)
        if not config.entry_price:
            open_order_price_type = PriceType.BestBid if config.side == TradeType.BUY else PriceType.BestAsk
            config.entry_price = self.get_price(config.connector_name, config.trading_pair,
                                                price_type=open_order_price_type)
        self.config: PositionExecutorConfig = config
        self.trading_rules = self.get_trading_rules(self.config.connector_name, self.config.trading_pair)

        # Order tracking
        self._open_order: Optional[TrackedOrder] = None
        self._close_order: Optional[TrackedOrder] = None
        self._take_profit_limit_order: Optional[TrackedOrder] = None
        self._failed_orders: List[TrackedOrder] = []
        self._trailing_stop_trigger_pct: Optional[Decimal] = None

        self._total_executed_amount_backup: Decimal = Decimal("0")
        self._current_retries = 0
        self._max_retries = max_retries

    @property
    def is_perpetual(self) -> bool:
        """
        Check if the exchange connector is perpetual.

        :return: True if the exchange connector is perpetual, False otherwise.
        """
        return self.is_perpetual_connector(self.config.connector_name)

    @property
    def is_trading(self):
        """
        Check if the position is trading.

        :return: True if the position is trading, False otherwise.
        """
        return self.status == RunnableStatus.RUNNING and self.open_filled_amount > Decimal("0")

    @property
    def open_filled_amount(self) -> Decimal:
        """
        Get the filled amount of the open order.

        :return: The filled amount of the open order if it exists, otherwise 0.
        """
        if self._open_order:
            if self._open_order.fee_asset == self.config.trading_pair.split("-")[0]:
                open_filled_amount = self._open_order.executed_amount_base - self._open_order.cum_fees_base
            else:
                open_filled_amount = self._open_order.executed_amount_base
            return self.connectors[self.config.connector_name].quantize_order_amount(
                trading_pair=self.config.trading_pair,
                amount=open_filled_amount)
        else:
            return Decimal("0")

    @property
    def amount_to_close(self) -> Decimal:
        """
        Get the amount to close the position.

        :return: The amount to close the position.
        """
        return self.open_filled_amount - self.close_filled_amount

    @property
    def open_filled_amount_quote(self) -> Decimal:
        """
        Get the filled amount of the open order in quote currency.

        :return: The filled amount of the open order in quote currency.
        """
        return self.open_filled_amount * self.entry_price

    @property
    def close_filled_amount(self) -> Decimal:
        """
        Get the filled amount of the close order.

        :return: The filled amount of the close order if it exists, otherwise 0.
        """
        return self._close_order.executed_amount_base if self._close_order else Decimal("0")

    @property
    def close_filled_amount_quote(self) -> Decimal:
        """
        Get the filled amount of the close order in quote currency.

        :return: The filled amount of the close order in quote currency.
        """
        return self.close_filled_amount * self.close_price

    @property
    def filled_amount(self) -> Decimal:
        """
        Get the filled amount of the position.
        """
        return self.open_filled_amount + self.close_filled_amount

    @property
    def filled_amount_quote(self) -> Decimal:
        """
        Get the filled amount of the position in quote currency.
        """
        return self.open_filled_amount_quote + self.close_filled_amount_quote if self.close_type != CloseType.POSITION_HOLD else Decimal("0")

    @property
    def is_expired(self) -> bool:
        """
        Check if the position is expired.

        :return: True if the position is expired, False otherwise.
        """
        return self.end_time and self.end_time <= self._strategy.current_timestamp

    @property
    def current_market_price(self) -> Decimal:
        """
        This method is responsible for getting the current market price to be used as a reference for control barriers.

        :return: The current market price.
        """
        price_type = PriceType.BestBid if self.config.side == TradeType.BUY else PriceType.BestAsk
        return self.get_price(self.config.connector_name, self.config.trading_pair, price_type=price_type)

    @property
    def entry_price(self) -> Decimal:
        """
        This method is responsible for getting the entry price. If the open order is done, it returns the average executed price.
        If the entry price is set in the configuration, it returns the entry price from the configuration.
        Otherwise, it returns the best ask price for buy orders and the best bid price for sell orders.

        :return: The entry price.
        """
        if self._open_order and self._open_order.is_done:
            return self._open_order.average_executed_price
        elif self.config.triple_barrier_config.open_order_type == OrderType.LIMIT_MAKER:
            if self.config.side == TradeType.BUY:
                best_bid = self.get_price(self.config.connector_name, self.config.trading_pair, PriceType.BestBid)
                return min(self.config.entry_price, best_bid)
            else:
                best_ask = self.get_price(self.config.connector_name, self.config.trading_pair, PriceType.BestAsk)
                return max(self.config.entry_price, best_ask)
        else:
            return self.config.entry_price

    @property
    def close_price(self) -> Decimal:
        """
        This method is responsible for getting the close price. If the close order is done, it returns the average executed price.
        Otherwise, it returns the current market price.

        :return: The close price.
        """
        if self._close_order and self._close_order.is_done:
            return self._close_order.average_executed_price
        else:
            return self.current_market_price

    @property
    def close_order_side(self):
        return TradeType.BUY if self.config.side == TradeType.SELL else TradeType.SELL

    @property
    def trade_pnl_pct(self) -> Decimal:
        """
        Calculate the trade pnl (Pure pnl without fees)

        :return: The trade pnl percentage.
        """
        if self.open_filled_amount != Decimal("0") and self.close_type not in [CloseType.FAILED, CloseType.POSITION_HOLD]:
            if self.config.side == TradeType.BUY:
                return (self.close_price - self.entry_price) / self.entry_price
            else:
                return (self.entry_price - self.close_price) / self.entry_price
        else:
            return Decimal("0")

    @property
    def trade_pnl_quote(self) -> Decimal:
        """
        Calculate the trade pnl in quote asset

        :return: The trade pnl in quote asset.
        """
        return self.trade_pnl_pct * self.open_filled_amount * self.entry_price

    def get_net_pnl_quote(self) -> Decimal:
        """
        Calculate the net pnl in quote asset

        :return: The net pnl in quote asset.
        """
        return self.trade_pnl_quote - self.cum_fees_quote

    def get_cum_fees_quote(self) -> Decimal:
        """
        Calculate the cumulative fees in quote asset

        :return: The cumulative fees in quote asset.
        """
        orders = [self._open_order, self._close_order]
        return sum([order.cum_fees_quote for order in orders if order])

    def get_net_pnl_pct(self) -> Decimal:
        """
        Calculate the net pnl percentage

        :return: The net pnl percentage.
        """
        return self.net_pnl_quote / self.open_filled_amount_quote if self.open_filled_amount_quote != Decimal("0") else Decimal("0")

    @property
    def end_time(self) -> Optional[float]:
        """
        Calculate the end time of the position based on the time limit

        :return: The end time of the position.
        """
        if not self.config.triple_barrier_config.time_limit:
            return None
        return self.config.timestamp + self.config.triple_barrier_config.time_limit

    @property
    def take_profit_price(self):
        """
        This method is responsible for calculating the take profit price to place the take profit limit order.

        :return: The take profit price.
        """
        if self.config.side == TradeType.BUY:
            take_profit_price = self.entry_price * (1 + self.config.triple_barrier_config.take_profit)
            if self.config.triple_barrier_config.take_profit_order_type == OrderType.LIMIT_MAKER:
                take_profit_price = max(take_profit_price,
                                        self.get_price(self.config.connector_name, self.config.trading_pair,
                                                       PriceType.BestAsk))
        else:
            take_profit_price = self.entry_price * (1 - self.config.triple_barrier_config.take_profit)
            if self.config.triple_barrier_config.take_profit_order_type == OrderType.LIMIT_MAKER:
                take_profit_price = min(take_profit_price,
                                        self.get_price(self.config.connector_name, self.config.trading_pair,
                                                       PriceType.BestBid))
        return take_profit_price

    async def control_task(self):
        """
        This method is responsible for controlling the task based on the status of the executor.

        :return: None
        """
        if self.status == RunnableStatus.RUNNING:
            self.control_open_order()
            self.control_barriers()
        elif self.status == RunnableStatus.SHUTTING_DOWN:
            await self.control_shutdown_process()
        self.evaluate_max_retries()

    def all_orders_completed(self):
        """
        This method is responsible for checking if the open orders are completed.

        :return: True if the open orders are completed, False otherwise.
        """
        open_order_condition = not self._open_order or self._open_order.is_done
        take_profit_condition = not self._take_profit_limit_order or self._take_profit_limit_order.is_done
        close_order_condition = not self._close_order or self._close_order.is_done
        return open_order_condition and take_profit_condition and close_order_condition

    async def control_shutdown_process(self):
        """
        This method is responsible for controlling the shutdown process of the executor.

        :return: None
        """
        self.close_timestamp = self._strategy.current_timestamp
        if self.all_orders_completed():
            if self.close_type == CloseType.POSITION_HOLD:
                if self._open_order and self._open_order.is_filled:
                    self._held_position_orders.append(self._open_order.order.to_json())
                if self._close_order and self._close_order.is_filled:
                    self._held_position_orders.append(self._close_order.order.to_json())
                if len(self._held_position_orders) == 0:
                    self.close_type = CloseType.EARLY_STOP
                self.stop()
            elif self.open_and_close_volume_match():
                self.stop()
            else:
                await self.control_close_order()
                self._current_retries += 1
        else:
            self.cancel_open_orders()
        await self._sleep(5.0)

    def open_and_close_volume_match(self):
        if self.open_filled_amount == Decimal("0"):
            return True
        else:
            return self._close_order and self._close_order.is_filled

    async def control_close_order(self):
        """
        This method is responsible for controlling the close order. If the close order is filled and the open orders are
        completed, it stops the executor. If the close order is not placed, it places the close order. If the close order
        is not filled, it waits for the close order to be filled and requests the order information to the connector.
        """
        if self._close_order:
            in_flight_order = self.get_in_flight_order(self.config.connector_name,
                                                       self._close_order.order_id) if not self._close_order.order else self._close_order.order
            if in_flight_order:
                self._close_order.order = in_flight_order
                connector = self.connectors[self.config.connector_name]
                await connector._update_orders_with_error_handler(
                    orders=[in_flight_order],
                    error_handler=connector._handle_update_error_for_lost_order)
                self.logger().info("Waiting for close order to be filled")
            else:
                self._failed_orders.append(self._close_order)
                self._close_order = None
        else:
            self.place_close_order_and_cancel_open_orders(close_type=self.close_type)

    def evaluate_max_retries(self):
        """
        This method is responsible for evaluating the maximum number of retries to place an order and stop the executor
        if the maximum number of retries is reached.

        :return: None
        """
        if self._current_retries > self._max_retries:
            self.close_type = CloseType.FAILED
            self.stop()

    async def on_start(self):
        """
        This method is responsible for starting the executor and validating if the position is expired. The base method
        validates if there is enough balance to place the open order.

        :return: None
        """
        await super().on_start()
        if self.is_expired:
            self.close_type = CloseType.EXPIRED
            self.stop()

    def control_open_order(self):
        """
        This method is responsible for controlling the open order. It checks if the open order is not placed and if the
        close price is within the activation bounds to place the open order.

        :return: None
        """
        if not self._open_order:
            if self._is_within_activation_bounds(self.config.entry_price, self.config.side,
                                                 self.config.triple_barrier_config.open_order_type):
                self.place_open_order()
        else:
            if self._open_order.order and not self._open_order.is_filled and \
                    not self._is_within_activation_bounds(self.config.entry_price, self.config.side,
                                                          self.config.triple_barrier_config.open_order_type):
                self.cancel_open_order()

    def _is_within_activation_bounds(self, order_price: Decimal, side: TradeType, order_type: OrderType) -> bool:
        """
        This method is responsible for checking if the close price is within the activation bounds to place the open
        order. If the activation bounds are not set, it returns True. This makes the executor more capital efficient.

        :param close_price: The close price to be checked.
        :return: True if the close price is within the activation bounds, False otherwise.
        """
        activation_bounds = self.config.activation_bounds
        mid_price = self.get_price(self.config.connector_name, self.config.trading_pair, PriceType.MidPrice)
        if activation_bounds:
            if order_type.is_limit_type():
                if side == TradeType.BUY:
                    return order_price >= mid_price * (1 - activation_bounds[0])
                else:
                    return order_price <= mid_price * (1 + activation_bounds[0])
            else:
                if side == TradeType.BUY:
                    min_price_to_buy = order_price * (1 - activation_bounds[0])
                    max_price_to_buy = order_price * (1 + activation_bounds[1])
                    return min_price_to_buy <= mid_price <= max_price_to_buy
                else:
                    min_price_to_sell = order_price * (1 - activation_bounds[1])
                    max_price_to_sell = order_price * (1 + activation_bounds[0])
                    return min_price_to_sell <= mid_price <= max_price_to_sell
        else:
            return True

    def place_open_order(self):
        """
        This method is responsible for placing the open order.

        :return: None
        """
        order_id = self.place_order(
            connector_name=self.config.connector_name,
            trading_pair=self.config.trading_pair,
            order_type=self.config.triple_barrier_config.open_order_type,
            amount=self.config.amount,
            price=self.entry_price,
            side=self.config.side,
            position_action=PositionAction.OPEN,
        )
        self._open_order = TrackedOrder(order_id=order_id)
        self.logger().debug(f"Executor ID: {self.config.id} - Placing open order {order_id}")

    def control_barriers(self):
        """
        This method is responsible for controlling the barriers. It controls the stop loss, take profit, time limit and
        trailing stop.

        :return: None
        """
        if self._open_order and self._open_order.is_filled and self.open_filled_amount >= self.trading_rules.min_order_size \
                and self.open_filled_amount_quote >= self.trading_rules.min_notional_size:
            self.control_stop_loss()
            self.control_trailing_stop()
            self.control_take_profit()
        self.control_time_limit()

    def place_close_order_and_cancel_open_orders(self, close_type: CloseType, price: Decimal = Decimal("NaN")):
        """
        This method is responsible for placing the close order and canceling the open orders. If the difference between
        the open filled amount and the close filled amount is greater than the minimum order size, it places the close
        order. It also cancels the open orders.

        :param close_type: The type of the close order.
        :param price: The price to be used in the close order.
        :return: None
        """
        self.cancel_open_orders()
        if self.amount_to_close >= self.trading_rules.min_order_size and close_type != CloseType.POSITION_HOLD:
            order_id = self.place_order(
                connector_name=self.config.connector_name,
                trading_pair=self.config.trading_pair,
                order_type=OrderType.MARKET,
                amount=self.amount_to_close,
                price=price,
                side=self.close_order_side,
                position_action=PositionAction.CLOSE,
            )
            self._close_order = TrackedOrder(order_id=order_id)
            self.logger().debug(f"Executor ID: {self.config.id} - Placing close order {order_id} --> Filled amount: {self.open_filled_amount}")
        self.close_type = close_type
        self.close_timestamp = self._strategy.current_timestamp
        self._status = RunnableStatus.SHUTTING_DOWN

    def cancel_open_orders(self):
        """
        This method is responsible for canceling the open orders.

        :return: None
        """
        if self._open_order and self._open_order.order and self._open_order.order.is_open:
            self.cancel_open_order()
        if self._take_profit_limit_order and self._take_profit_limit_order.order and self._take_profit_limit_order.order.is_open:
            self.cancel_take_profit()

    def control_stop_loss(self):
        """
        This method is responsible for controlling the stop loss. If the net pnl percentage is less than the stop loss
        percentage, it places the close order and cancels the open orders.

        :return: None
        """
        if self.config.triple_barrier_config.stop_loss:
            if self.net_pnl_pct <= -self.config.triple_barrier_config.stop_loss:
                self.place_close_order_and_cancel_open_orders(close_type=CloseType.STOP_LOSS)

    def control_take_profit(self):
        """
        This method is responsible for controlling the take profit. If the net pnl percentage is greater than the take
        profit percentage, it places the close order and cancels the open orders. If the take profit order type is limit,
        it places the take profit limit order. If the amount of the take profit order is different than the total amount
        executed in the open order, it renews the take profit order (can happen with partial fills).

        :return: None
        """
        if self.config.triple_barrier_config.take_profit:
            if self.config.triple_barrier_config.take_profit_order_type.is_limit_type():
                is_within_activation_bounds = self._is_within_activation_bounds(
                    self.take_profit_price, self.close_order_side,
                    self.config.triple_barrier_config.take_profit_order_type)
                if not self._take_profit_limit_order:
                    if is_within_activation_bounds:
                        self.place_take_profit_limit_order()
                else:
                    if self._take_profit_limit_order.is_open and not self._take_profit_limit_order.is_filled and \
                            not is_within_activation_bounds:
                        self.cancel_take_profit()
            elif self.net_pnl_pct >= self.config.triple_barrier_config.take_profit:
                self.place_close_order_and_cancel_open_orders(close_type=CloseType.TAKE_PROFIT)

    def control_time_limit(self):
        """
        This method is responsible for controlling the time limit. If the position is expired, it places the close order
        and cancels the open orders.

        :return: None
        """
        if self.is_expired:
            self.place_close_order_and_cancel_open_orders(close_type=CloseType.TIME_LIMIT)

    def place_take_profit_limit_order(self):
        """
        This method is responsible for placing the take profit limit order.

        :return: None
        """
        order_id = self.place_order(
            connector_name=self.config.connector_name,
            trading_pair=self.config.trading_pair,
            amount=self.amount_to_close,
            price=self.take_profit_price,
            order_type=self.config.triple_barrier_config.take_profit_order_type,
            position_action=PositionAction.CLOSE,
            side=self.close_order_side,
        )
        self._take_profit_limit_order = TrackedOrder(order_id=order_id)
        self.logger().debug(f"Executor ID: {self.config.id} - Placing take profit order {order_id}")

    def renew_take_profit_order(self):
        """
        This method is responsible for renewing the take profit order.

        :return: None
        """
        self.cancel_take_profit()
        self.place_take_profit_limit_order()
        self.logger().debug("Renewing take profit order")

    def cancel_take_profit(self):
        """
        This method is responsible for canceling the take profit order.

        :return: None
        """
        self._strategy.cancel(
            connector_name=self.config.connector_name,
            trading_pair=self.config.trading_pair,
            order_id=self._take_profit_limit_order.order_id
        )
        self.logger().debug("Removing take profit")

    def cancel_open_order(self):
        """
        This method is responsible for canceling the open order.

        :return: None
        """
        self._strategy.cancel(
            connector_name=self.config.connector_name,
            trading_pair=self.config.trading_pair,
            order_id=self._open_order.order_id
        )
        self.logger().debug("Removing open order")

    def early_stop(self, keep_position: bool = False):
        """
        This method allows strategy to stop the executor early.

        :return: None
        """
        self.close_type = CloseType.POSITION_HOLD if keep_position else CloseType.EARLY_STOP
        self._status = RunnableStatus.SHUTTING_DOWN

    def update_tracked_orders_with_order_id(self, order_id: str):
        """
        This method is responsible for updating the tracked orders with the information from the InFlightOrder, using
        the order_id as a reference.

        :param order_id: The order_id to be used as a reference.
        :return: None
        """
        in_flight_order = self.get_in_flight_order(self.config.connector_name, order_id)
        if self._open_order and self._open_order.order_id == order_id:
            self._open_order.order = in_flight_order
        elif self._close_order and self._close_order.order_id == order_id:
            self._close_order.order = in_flight_order
        elif self._take_profit_limit_order and self._take_profit_limit_order.order_id == order_id:
            self._take_profit_limit_order.order = in_flight_order

    def process_order_created_event(self, _, market, event: Union[BuyOrderCreatedEvent, SellOrderCreatedEvent]):
        """
        This method is responsible for processing the order created event. Here we will update the TrackedOrder with the
        order_id.
        """
        self.update_tracked_orders_with_order_id(event.order_id)

    def process_order_filled_event(self, _, market, event: OrderFilledEvent):
        """
        This method is responsible for processing the order filled event. Here we will update the value of
        _total_executed_amount_backup, that can be used if the InFlightOrder
        is not available.
        """
        self.update_tracked_orders_with_order_id(event.order_id)

    def process_order_completed_event(self, _, market, event: Union[BuyOrderCompletedEvent, SellOrderCompletedEvent]):
        """
        This method is responsible for processing the order completed event. Here we will check if the id is one of the
        tracked orders and update the state
        """
        self._total_executed_amount_backup += event.base_asset_amount
        self.update_tracked_orders_with_order_id(event.order_id)

        if self._take_profit_limit_order and self._take_profit_limit_order.order_id == event.order_id:
            self.close_type = CloseType.TAKE_PROFIT
            self._close_order = self._take_profit_limit_order
            self._status = RunnableStatus.SHUTTING_DOWN

    def process_order_canceled_event(self, _, market: ConnectorBase, event: OrderCancelledEvent):
        """
        This method is responsible for processing the order canceled event
        """
        if self._close_order and event.order_id == self._close_order.order_id:
            self._failed_orders.append(self._close_order)
            self._close_order = None
        if self._open_order and event.order_id == self._open_order.order_id:
            self._failed_orders.append(self._open_order)
            self._open_order = None
        if self._take_profit_limit_order and event.order_id == self._take_profit_limit_order.order_id:
            self._failed_orders.append(self._take_profit_limit_order)
            self._take_profit_limit_order = None

    def process_order_failed_event(self, _, market, event: MarketOrderFailureEvent):
        """
        This method is responsible for processing the order failed event. Here we will add the InFlightOrder to the
        failed orders list.
        """
        if self._open_order and event.order_id == self._open_order.order_id:
            self._failed_orders.append(self._open_order)
            self._open_order = None
            self.logger().error(f"Open order failed {event.order_id}. Retrying {self._current_retries}/{self._max_retries}")
            self._current_retries += 1
        elif self._close_order and event.order_id == self._close_order.order_id:
            self._failed_orders.append(self._close_order)
            self._close_order = None
            self.logger().error(f"Close order failed {event.order_id}. Retrying {self._current_retries}/{self._max_retries}")
            self._current_retries += 1
        elif self._take_profit_limit_order and event.order_id == self._take_profit_limit_order.order_id:
            self._failed_orders.append(self._take_profit_limit_order)
            self._take_profit_limit_order = None
            self.logger().error(f"Take profit order failed {event.order_id}. Retrying {self._current_retries}/{self._max_retries}")

    def get_custom_info(self) -> Dict:
        return {
            "level_id": self.config.level_id,
            "current_position_average_price": self.entry_price,
            "side": self.config.side,
            "current_retries": self._current_retries,
            "max_retries": self._max_retries,
            "close_price": self.close_price,
            "open_order_last_update": self._open_order.last_update_timestamp if self._open_order else None,
            "order_ids": [order.order_id for order in [self._open_order, self._close_order, self._take_profit_limit_order] if order],
            "held_position_orders": self._held_position_orders,
        }

    def to_format_status(self, scale=1.0):
        lines = []
        current_price = self.get_price(self.config.connector_name, self.config.trading_pair)
        amount_in_quote = self.entry_price * (self.open_filled_amount if self.open_filled_amount > Decimal("0") else self.config.amount)
        quote_asset = self.config.trading_pair.split("-")[1]
        if self.is_closed:
            lines.extend([f"""
| Trading Pair: {self.config.trading_pair} | Exchange: {self.config.connector_name} | Side: {self.config.side}
| Entry price: {self.entry_price:.6f} | Close price: {self.close_price:.6f} | Amount: {amount_in_quote:.4f} {quote_asset}
| Realized PNL: {self.trade_pnl_quote:.6f} {quote_asset} | Total Fee: {self.cum_fees_quote:.6f} {quote_asset}
| PNL (%): {self.net_pnl_pct * 100:.2f}% | PNL (abs): {self.net_pnl_quote:.6f} {quote_asset} | Close Type: {self.close_type}
"""])
        else:
            lines.extend([f"""
| Trading Pair: {self.config.trading_pair} | Exchange: {self.config.connector_name} | Side: {self.config.side} |
| Entry price: {self.entry_price:.6f} | Close price: {self.close_price:.6f} | Amount: {amount_in_quote:.4f} {quote_asset}
| Unrealized PNL: {self.trade_pnl_quote:.6f} {quote_asset} | Total Fee: {self.cum_fees_quote:.6f} {quote_asset}
| PNL (%): {self.net_pnl_pct * 100:.2f}% | PNL (abs): {self.net_pnl_quote:.6f} {quote_asset} | Close Type: {self.close_type}
        """])

        if self.is_trading:
            progress = 0
            if self.config.triple_barrier_config.time_limit:
                time_scale = int(scale * 60)
                seconds_remaining = (self.end_time - self._strategy.current_timestamp)
                time_progress = (self.config.triple_barrier_config.time_limit - seconds_remaining) / self.config.triple_barrier_config.time_limit
                time_bar = "".join(['*' if i < time_scale * time_progress else '-' for i in range(time_scale)])
                lines.extend([f"Time limit: {time_bar}"])

            if self.config.triple_barrier_config.take_profit and self.config.triple_barrier_config.stop_loss:
                price_scale = int(scale * 60)
                stop_loss_price = self.entry_price * (1 - self.config.triple_barrier_config.stop_loss) if self.config.side == TradeType.BUY \
                    else self.entry_price * (1 + self.config.triple_barrier_config.stop_loss)
                take_profit_price = self.entry_price * (1 + self.config.triple_barrier_config.take_profit) if self.config.side == TradeType.BUY \
                    else self.entry_price * (1 - self.config.triple_barrier_config.take_profit)
                if self.config.side == TradeType.BUY:
                    price_range = take_profit_price - stop_loss_price
                    progress = (current_price - stop_loss_price) / price_range
                elif self.config.side == TradeType.SELL:
                    price_range = stop_loss_price - take_profit_price
                    progress = (stop_loss_price - current_price) / price_range
                price_bar = [f'--{current_price:.5f}--' if i == int(price_scale * progress) else '-' for i in range(price_scale)]
                price_bar.insert(0, f"SL:{stop_loss_price:.5f}")
                price_bar.append(f"TP:{take_profit_price:.5f}")
                lines.extend(["".join(price_bar)])
            if self.config.triple_barrier_config.trailing_stop:
                lines.extend([f"Trailing stop pnl trigger: {self._trailing_stop_trigger_pct:.5f}"])
            lines.extend(["-----------------------------------------------------------------------------------------------------------"])
        return lines

    def control_trailing_stop(self):
        if self.config.triple_barrier_config.trailing_stop:
            net_pnl_pct = self.get_net_pnl_pct()
            if not self._trailing_stop_trigger_pct:
                if net_pnl_pct > self.config.triple_barrier_config.trailing_stop.activation_price:
                    self._trailing_stop_trigger_pct = net_pnl_pct - self.config.triple_barrier_config.trailing_stop.trailing_delta
            else:
                if net_pnl_pct < self._trailing_stop_trigger_pct:
                    self.place_close_order_and_cancel_open_orders(close_type=CloseType.TRAILING_STOP)
                if net_pnl_pct - self.config.triple_barrier_config.trailing_stop.trailing_delta > self._trailing_stop_trigger_pct:
                    self._trailing_stop_trigger_pct = net_pnl_pct - self.config.triple_barrier_config.trailing_stop.trailing_delta

    async def validate_sufficient_balance(self):
        if self.is_perpetual:
            order_candidate = PerpetualOrderCandidate(
                trading_pair=self.config.trading_pair,
                is_maker=self.config.triple_barrier_config.open_order_type.is_limit_type(),
                order_type=self.config.triple_barrier_config.open_order_type,
                order_side=self.config.side,
                amount=self.config.amount,
                price=self.entry_price,
                leverage=Decimal(self.config.leverage),
            )
        else:
            order_candidate = OrderCandidate(
                trading_pair=self.config.trading_pair,
                is_maker=self.config.triple_barrier_config.open_order_type.is_limit_type(),
                order_type=self.config.triple_barrier_config.open_order_type,
                order_side=self.config.side,
                amount=self.config.amount,
                price=self.entry_price,
            )
        adjusted_order_candidates = self.adjust_order_candidates(self.config.connector_name, [order_candidate])
        if adjusted_order_candidates[0].amount == Decimal("0"):
            self.close_type = CloseType.INSUFFICIENT_BALANCE
            self.logger().error("Not enough budget to open position.")
            self.stop()

    async def _sleep(self, delay: float):
        """
        This method is responsible for sleeping the executor for a specific time.

        :param delay: The time to sleep.
        :return: None
        """
        await asyncio.sleep(delay)



import asyncio
from decimal import Decimal
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Field

from hummingbot.connector.gateway.common_types import ConnectorType, get_connector_type
from hummingbot.connector.gateway.gateway_in_flight_order import GatewayInFlightOrder
from hummingbot.connector.gateway.gateway_swap import GatewaySwap
from hummingbot.core.data_type.common import LPType, OrderType, TradeType
from hummingbot.core.data_type.trade_fee import TokenAmount, TradeFeeBase
from hummingbot.core.event.events import (
    MarketEvent,
    RangePositionLiquidityAddedEvent,
    RangePositionLiquidityRemovedEvent,
    RangePositionUpdateFailureEvent,
)
from hummingbot.core.utils import async_ttl_cache
from hummingbot.core.utils.async_utils import safe_ensure_future


class TokenInfo(BaseModel):
    address: str
    symbol: str
    decimals: int


class AMMPoolInfo(BaseModel):
    address: str
    base_token_address: str = Field(alias="baseTokenAddress")
    quote_token_address: str = Field(alias="quoteTokenAddress")
    price: float
    fee_pct: float = Field(alias="feePct")
    base_token_amount: float = Field(alias="baseTokenAmount")
    quote_token_amount: float = Field(alias="quoteTokenAmount")


class CLMMPoolInfo(BaseModel):
    address: str
    base_token_address: str = Field(alias="baseTokenAddress")
    quote_token_address: str = Field(alias="quoteTokenAddress")
    bin_step: int = Field(alias="binStep")
    fee_pct: float = Field(alias="feePct")
    price: float
    base_token_amount: float = Field(alias="baseTokenAmount")
    quote_token_amount: float = Field(alias="quoteTokenAmount")
    active_bin_id: int = Field(alias="activeBinId")


class AMMPositionInfo(BaseModel):
    pool_address: str = Field(alias="poolAddress")
    wallet_address: str = Field(alias="walletAddress")
    base_token_address: str = Field(alias="baseTokenAddress")
    quote_token_address: str = Field(alias="quoteTokenAddress")
    lp_token_amount: float = Field(alias="lpTokenAmount")
    base_token_amount: float = Field(alias="baseTokenAmount")
    quote_token_amount: float = Field(alias="quoteTokenAmount")
    price: float
    base_token: Optional[str] = None
    quote_token: Optional[str] = None


class CLMMPositionInfo(BaseModel):
    address: str
    pool_address: str = Field(alias="poolAddress")
    base_token_address: str = Field(alias="baseTokenAddress")
    quote_token_address: str = Field(alias="quoteTokenAddress")
    base_token_amount: float = Field(alias="baseTokenAmount")
    quote_token_amount: float = Field(alias="quoteTokenAmount")
    base_fee_amount: float = Field(alias="baseFeeAmount")
    quote_fee_amount: float = Field(alias="quoteFeeAmount")
    lower_bin_id: int = Field(alias="lowerBinId")
    upper_bin_id: int = Field(alias="upperBinId")
    lower_price: float = Field(alias="lowerPrice")
    upper_price: float = Field(alias="upperPrice")
    price: float
    base_token: Optional[str] = None
    quote_token: Optional[str] = None


class GatewayLp(GatewaySwap):
    """
    Handles AMM and CLMM liquidity provision functionality including fetching pool info and adding/removing liquidity.
    Maintains order tracking and wallet interactions in the base class.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Store LP operation metadata for triggering proper events
        self._lp_orders_metadata: Dict[str, Dict] = {}

    def _trigger_lp_events_if_needed(self, order_id: str, transaction_hash: str):
        """
        Helper to trigger LP-specific events when an order completes.
        This is called by both fast monitoring and slow polling to avoid duplication.
        """
        # Check if already triggered (metadata would be deleted)
        if order_id not in self._lp_orders_metadata:
            return

        tracked_order = self._order_tracker.fetch_order(order_id)
        if not tracked_order or tracked_order.trade_type != TradeType.RANGE:
            return

        metadata = self._lp_orders_metadata[order_id]

        # Trigger appropriate event based on transaction result
        # For LP operations (RANGE orders), state stays OPEN even when done, so check is_done
        is_successful = tracked_order.is_done and not tracked_order.is_failure and not tracked_order.is_cancelled

        if is_successful:
            # Transaction successful - trigger LP-specific events
            if metadata["operation"] == "add":
                self._trigger_add_liquidity_event(
                    order_id=order_id,
                    exchange_order_id=transaction_hash,
                    trading_pair=tracked_order.trading_pair,
                    lower_price=metadata["lower_price"],
                    upper_price=metadata["upper_price"],
                    amount=metadata["amount"],
                    fee_tier=metadata["fee_tier"],
                    creation_timestamp=tracked_order.creation_timestamp,
                    trade_fee=TradeFeeBase.new_spot_fee(
                        fee_schema=self.trade_fee_schema(),
                        trade_type=tracked_order.trade_type,
                        flat_fees=[TokenAmount(amount=metadata.get("tx_fee", Decimal("0")), token=self._native_currency)]
                    ),
                    # P&L tracking fields from gateway response
                    position_address=metadata.get("position_address", ""),
                    base_amount=metadata.get("base_amount", Decimal("0")),
                    quote_amount=metadata.get("quote_amount", Decimal("0")),
                    position_rent=metadata.get("position_rent", Decimal("0")),
                )
            elif metadata["operation"] == "remove":
                self._trigger_remove_liquidity_event(
                    order_id=order_id,
                    exchange_order_id=transaction_hash,
                    trading_pair=tracked_order.trading_pair,
                    token_id=metadata["position_address"],
                    creation_timestamp=tracked_order.creation_timestamp,
                    trade_fee=TradeFeeBase.new_spot_fee(
                        fee_schema=self.trade_fee_schema(),
                        trade_type=tracked_order.trade_type,
                        flat_fees=[TokenAmount(amount=metadata.get("tx_fee", Decimal("0")), token=self._native_currency)]
                    ),
                    # P&L tracking fields from gateway response
                    position_address=metadata.get("position_address", ""),
                    base_amount=metadata.get("base_amount", Decimal("0")),
                    quote_amount=metadata.get("quote_amount", Decimal("0")),
                    base_fee=metadata.get("base_fee", Decimal("0")),
                    quote_fee=metadata.get("quote_fee", Decimal("0")),
                    position_rent_refunded=metadata.get("position_rent_refunded", Decimal("0")),
                )
        elif tracked_order.is_failure:
            # Transaction failed - trigger LP-specific failure event for strategy handling
            operation_type = "add" if metadata["operation"] == "add" else "remove"
            self.logger().error(
                f"LP {operation_type} liquidity transaction failed for order {order_id} (tx: {transaction_hash})"
            )
            # Trigger RangePositionUpdateFailureEvent so strategies can retry
            self.trigger_event(
                MarketEvent.RangePositionUpdateFailure,
                RangePositionUpdateFailureEvent(
                    timestamp=self.current_timestamp,
                    order_id=order_id,
                    order_action=LPType.ADD if metadata["operation"] == "add" else LPType.REMOVE,
                )
            )
        elif tracked_order.is_cancelled:
            # Transaction cancelled
            operation_type = "add" if metadata["operation"] == "add" else "remove"
            self.logger().warning(
                f"LP {operation_type} liquidity transaction cancelled for order {order_id} (tx: {transaction_hash})"
            )

        # Clean up metadata (prevents double-triggering) and stop tracking
        del self._lp_orders_metadata[order_id]
        self.stop_tracking_order(order_id)

    async def update_order_status(self, tracked_orders: List[GatewayInFlightOrder]):
        """
        Override to trigger RangePosition events after LP transactions complete (batch polling).
        """
        # Call parent implementation (handles timeout checking)
        await super().update_order_status(tracked_orders)

        # Trigger LP events for any completed LP operations
        for tracked_order in tracked_orders:
            if tracked_order.trade_type == TradeType.RANGE:
                # Get transaction hash
                try:
                    tx_hash = await tracked_order.get_exchange_order_id()
                    self._trigger_lp_events_if_needed(tracked_order.client_order_id, tx_hash)
                except Exception as e:
                    self.logger().warning(f"Error triggering LP event for {tracked_order.client_order_id}: {e}", exc_info=True)

    # Error code from gateway for transaction confirmation timeout
    TRANSACTION_TIMEOUT_CODE = "TRANSACTION_TIMEOUT"

    def _handle_operation_failure(self, order_id: str, trading_pair: str, operation_name: str, error: Exception):
        """
        Override to trigger RangePositionUpdateFailureEvent for LP operations.
        Only triggers retry for transaction confirmation timeouts (code: TRANSACTION_TIMEOUT).
        """
        # Call parent implementation
        super()._handle_operation_failure(order_id, trading_pair, operation_name, error)

        # Check if this is a transaction timeout error (retryable)
        # Gateway returns error with code "TRANSACTION_TIMEOUT" for tx confirmation timeouts
        error_str = str(error)
        is_timeout_error = self.TRANSACTION_TIMEOUT_CODE in error_str

        if is_timeout_error and order_id in self._lp_orders_metadata:
            metadata = self._lp_orders_metadata[order_id]
            operation = metadata.get("operation", "")
            self.logger().warning(
                f"Transaction timeout detected for LP {operation} order {order_id} on {trading_pair}. "
                f"Chain may be congested. Triggering retry event..."
            )
            self.trigger_event(
                MarketEvent.RangePositionUpdateFailure,
                RangePositionUpdateFailureEvent(
                    timestamp=self.current_timestamp,
                    order_id=order_id,
                    order_action=LPType.ADD if operation == "add" else LPType.REMOVE,
                )
            )
            # Clean up metadata
            del self._lp_orders_metadata[order_id]
        elif order_id in self._lp_orders_metadata:
            # Non-retryable error, just clean up metadata
            self.logger().warning(f"Non-retryable error for {order_id}: {error_str[:100]}")
            del self._lp_orders_metadata[order_id]

    def _trigger_add_liquidity_event(
        self,
        order_id: str,
        exchange_order_id: str,
        trading_pair: str,
        lower_price: Decimal,
        upper_price: Decimal,
        amount: Decimal,
        fee_tier: str,
        creation_timestamp: float,
        trade_fee: TradeFeeBase,
        position_address: str = "",
        base_amount: Decimal = Decimal("0"),
        quote_amount: Decimal = Decimal("0"),
        mid_price: Decimal = Decimal("0"),
        position_rent: Decimal = Decimal("0"),
    ):
        """Trigger RangePositionLiquidityAddedEvent"""
        event = RangePositionLiquidityAddedEvent(
            timestamp=self.current_timestamp,
            order_id=order_id,
            exchange_order_id=exchange_order_id,
            trading_pair=trading_pair,
            lower_price=lower_price,
            upper_price=upper_price,
            amount=amount,
            fee_tier=fee_tier,
            creation_timestamp=creation_timestamp,
            trade_fee=trade_fee,
            token_id=0,
            # P&L tracking fields
            position_address=position_address,
            mid_price=mid_price,
            base_amount=base_amount,
            quote_amount=quote_amount,
            position_rent=position_rent,
        )
        self.trigger_event(MarketEvent.RangePositionLiquidityAdded, event)
        self.logger().info(f"Triggered RangePositionLiquidityAddedEvent for order {order_id}")

    def _trigger_remove_liquidity_event(
        self,
        order_id: str,
        exchange_order_id: str,
        trading_pair: str,
        token_id: str,
        creation_timestamp: float,
        trade_fee: TradeFeeBase,
        position_address: str = "",
        lower_price: Decimal = Decimal("0"),
        upper_price: Decimal = Decimal("0"),
        mid_price: Decimal = Decimal("0"),
        base_amount: Decimal = Decimal("0"),
        quote_amount: Decimal = Decimal("0"),
        base_fee: Decimal = Decimal("0"),
        quote_fee: Decimal = Decimal("0"),
        position_rent_refunded: Decimal = Decimal("0"),
    ):
        """Trigger RangePositionLiquidityRemovedEvent"""
        event = RangePositionLiquidityRemovedEvent(
            timestamp=self.current_timestamp,
            order_id=order_id,
            exchange_order_id=exchange_order_id,
            trading_pair=trading_pair,
            token_id=token_id,
            trade_fee=trade_fee,
            creation_timestamp=creation_timestamp,
            # P&L tracking fields
            position_address=position_address,
            lower_price=lower_price,
            upper_price=upper_price,
            mid_price=mid_price,
            base_amount=base_amount,
            quote_amount=quote_amount,
            base_fee=base_fee,
            quote_fee=quote_fee,
            position_rent_refunded=position_rent_refunded,
        )
        self.trigger_event(MarketEvent.RangePositionLiquidityRemoved, event)
        self.logger().info(f"Triggered RangePositionLiquidityRemovedEvent for order {order_id}")

    @async_ttl_cache(ttl=300, maxsize=10)
    async def get_pool_address(self, trading_pair: str) -> Optional[str]:
        """Get pool address for a trading pair (cached for 5 minutes)"""
        try:
            # Parse connector to get type (amm or clmm)
            connector_type = get_connector_type(self.connector_name)
            pool_type = "clmm" if connector_type == ConnectorType.CLMM else "amm"

            # Get pool info from gateway using the get_pool method
            connector_name = self.connector_name.split("/")[0]
            pool_info = await self._get_gateway_instance().get_pool(
                trading_pair=trading_pair,
                connector=connector_name,
                network=self.network,
                type=pool_type
            )

            pool_address = pool_info.get("address")
            if not pool_address:
                self.logger().warning(f"No pool address found for {trading_pair}")

            return pool_address

        except Exception as e:
            self.logger().error(f"Error getting pool address for {trading_pair}: {e}")
            return None

    @async_ttl_cache(ttl=5, maxsize=10)
    async def get_pool_info_by_address(
        self,
        pool_address: str,
    ) -> Optional[Union[AMMPoolInfo, CLMMPoolInfo]]:
        """
        Retrieves pool information by pool address directly.
        Uses the appropriate model (AMMPoolInfo or CLMMPoolInfo) based on connector type.

        :param pool_address: The pool contract address
        :return: Pool info object or None if not found
        """
        try:
            resp: Dict[str, Any] = await self._get_gateway_instance().pool_info(
                connector=self.connector_name,
                network=self.network,
                pool_address=pool_address,
            )

            if not resp:
                return None

            # Determine which model to use based on connector type
            connector_type = get_connector_type(self.connector_name)
            if connector_type == ConnectorType.CLMM:
                return CLMMPoolInfo(**resp)
            elif connector_type == ConnectorType.AMM:
                return AMMPoolInfo(**resp)
            else:
                self.logger().warning(f"Unknown connector type: {connector_type} for {self.connector_name}")
                return None

        except asyncio.CancelledError:
            raise
        except Exception as e:
            self.logger().network(
                f"Error fetching pool info for address {pool_address} on {self.connector_name}.",
                exc_info=True,
                app_warning_msg=str(e)
            )
            return None

    async def resolve_trading_pair_from_pool(
        self,
        pool_address: str,
    ) -> Optional[Dict[str, str]]:
        """
        Resolve trading pair information from pool address.
        Fetches pool info and returns token symbols and addresses.

        :param pool_address: The pool contract address
        :return: Dictionary with trading_pair, base_token, quote_token, base_token_address, quote_token_address
                 or None if pool not found
        """
        try:
            # Fetch pool info
            pool_info_resp = await self._get_gateway_instance().pool_info(
                connector=self.connector_name,
                network=self.network,
                pool_address=pool_address
            )

            if not pool_info_resp:
                raise ValueError(f"Could not fetch pool info for pool address {pool_address}")

            # Get token addresses from pool info
            base_token_address = pool_info_resp.get("baseTokenAddress")
            quote_token_address = pool_info_resp.get("quoteTokenAddress")

            if not base_token_address or not quote_token_address:
                raise ValueError(f"Pool info missing token addresses: {pool_info_resp}")

            # Try to get token symbols from connector's token cache
            base_token_info = self.get_token_by_address(base_token_address)
            quote_token_info = self.get_token_by_address(quote_token_address)

            base_symbol = base_token_info.get("symbol") if base_token_info else base_token_address
            quote_symbol = quote_token_info.get("symbol") if quote_token_info else quote_token_address

            trading_pair = f"{base_symbol}-{quote_symbol}"

            return {
                "trading_pair": trading_pair,
                "base_token": base_symbol,
                "quote_token": quote_symbol,
                "base_token_address": base_token_address,
                "quote_token_address": quote_token_address,
            }

        except asyncio.CancelledError:
            raise
        except Exception as e:
            self.logger().error(f"Error resolving trading pair from pool {pool_address}: {str(e)}", exc_info=True)
            return None

    async def get_pool_info(
        self,
        trading_pair: str,
    ) -> Optional[Union[AMMPoolInfo, CLMMPoolInfo]]:
        """
        Retrieves pool information for a given trading pair.
        Uses the appropriate model (AMMPoolInfo or CLMMPoolInfo) based on connector type.
        """
        try:
            # First get the pool address for the trading pair
            pool_address = await self.get_pool_address(trading_pair)

            if not pool_address:
                self.logger().warning(f"Could not find pool address for {trading_pair}")
                return None

            return await self.get_pool_info_by_address(pool_address)

        except asyncio.CancelledError:
            raise
        except Exception as e:
            self.logger().network(
                f"Error fetching pool info for {trading_pair} on {self.connector_name}.",
                exc_info=True,
                app_warning_msg=str(e)
            )
            return None

    def add_liquidity(self, trading_pair: str, price: float, **request_args) -> str:
        """
        Adds liquidity to a pool - either concentrated (CLMM) or regular (AMM) based on the connector type.
        :param trading_pair: The market trading pair
        :param price: The center price for the position.
        :param request_args: Additional arguments for liquidity addition
        :return: A newly created order id (internal).
        """
        trade_type: TradeType = TradeType.RANGE
        order_id: str = self.create_market_order_id(trade_type, trading_pair)

        # Check connector type and call appropriate function
        connector_type = get_connector_type(self.connector_name)
        if connector_type == ConnectorType.CLMM:
            safe_ensure_future(self._clmm_add_liquidity(trade_type, order_id, trading_pair, price, **request_args))
        elif connector_type == ConnectorType.AMM:
            safe_ensure_future(self._amm_add_liquidity(trade_type, order_id, trading_pair, price, **request_args))
        else:
            raise ValueError(f"Connector type {connector_type} does not support liquidity provision")

        return order_id

    async def _clmm_add_liquidity(
        self,
        trade_type: TradeType,
        order_id: str,
        trading_pair: str,
        price: float,
        lower_price: Optional[float] = None,
        upper_price: Optional[float] = None,
        upper_width_pct: Optional[float] = None,
        lower_width_pct: Optional[float] = None,
        base_token_amount: Optional[float] = None,
        quote_token_amount: Optional[float] = None,
        slippage_pct: Optional[float] = None,
        pool_address: Optional[str] = None,
        extra_params: Optional[Dict[str, Any]] = None,
    ):
        """
        Opens a concentrated liquidity position with explicit price range or calculated from percentages.

        :param trade_type: The trade type (should be RANGE)
        :param order_id: Internal order id (also called client_order_id)
        :param trading_pair: The trading pair for the position
        :param price: The center price for the position (used if lower/upper_price not provided)
        :param lower_price: Explicit lower price bound (takes priority over percentages)
        :param upper_price: Explicit upper price bound (takes priority over percentages)
        :param upper_width_pct: The upper range width percentage from center price (e.g. 10.0 for +10%)
        :param lower_width_pct: The lower range width percentage from center price (e.g. 5.0 for -5%)
        :param base_token_amount: Amount of base token to add (optional)
        :param quote_token_amount: Amount of quote token to add (optional)
        :param slippage_pct: Maximum allowed slippage percentage
        :param pool_address: Explicit pool address (optional, will lookup by trading_pair if not provided)
        :param extra_params: Optional connector-specific parameters (e.g., {"strategyType": 0} for Meteora)
        :return: Response from the gateway API
        """
        # Check connector type is CLMM
        if get_connector_type(self.connector_name) != ConnectorType.CLMM:
            raise ValueError(f"Connector {self.connector_name} is not of type CLMM.")

        # Split trading_pair to get base and quote tokens
        tokens = trading_pair.split("-")
        if len(tokens) != 2:
            raise ValueError(f"Invalid trading pair format: {trading_pair}")

        base_token, quote_token = tokens

        # Calculate the total amount in base token units
        base_amount = base_token_amount or 0.0
        quote_amount_in_base = (quote_token_amount or 0.0) / price if price > 0 else 0.0
        total_amount_in_base = base_amount + quote_amount_in_base

        # Start tracking order with calculated amount
        self.start_tracking_order(order_id=order_id,
                                  trading_pair=trading_pair,
                                  trade_type=trade_type,
                                  price=Decimal(str(price)),
                                  amount=Decimal(str(total_amount_in_base)),
                                  order_type=OrderType.AMM_ADD)

        # Determine position price range
        # Priority: explicit prices > width percentages
        if lower_price is not None and upper_price is not None:
            # Use explicit price bounds (highest priority)
            pass  # lower_price and upper_price already set
        elif upper_width_pct is not None and lower_width_pct is not None:
            # Calculate from width percentages
            lower_width_decimal = lower_width_pct / 100.0
            upper_width_decimal = upper_width_pct / 100.0
            lower_price = price * (1 - lower_width_decimal)
            upper_price = price * (1 + upper_width_decimal)
        else:
            raise ValueError("Must provide either (lower_price and upper_price) or (upper_width_pct and lower_width_pct)")

        # Get pool address - use explicit if provided, otherwise lookup by trading pair
        if not pool_address:
            pool_address = await self.get_pool_address(trading_pair)
            if not pool_address:
                raise ValueError(f"Could not find pool for {trading_pair}")

        # Store metadata for event triggering (will be enriched with response data)
        self._lp_orders_metadata[order_id] = {
            "operation": "add",
            "lower_price": Decimal(str(lower_price)),
            "upper_price": Decimal(str(upper_price)),
            "amount": Decimal(str(total_amount_in_base)),
            "fee_tier": pool_address,  # Use pool address as fee tier identifier
        }

        # Open position
        try:
            transaction_result = await self._get_gateway_instance().clmm_open_position(
                connector=self.connector_name,
                network=self.network,
                wallet_address=self.address,
                pool_address=pool_address,
                lower_price=lower_price,
                upper_price=upper_price,
                base_token_amount=base_token_amount,
                quote_token_amount=quote_token_amount,
                slippage_pct=slippage_pct,
                extra_params=extra_params
            )
            transaction_hash: Optional[str] = transaction_result.get("signature")
            if transaction_hash is not None and transaction_hash != "":
                self.update_order_from_hash(order_id, trading_pair, transaction_hash, transaction_result)
                # Store response data in metadata for P&L tracking
                # Gateway returns positive values for token amounts
                data = transaction_result.get("data", {})
                self._lp_orders_metadata[order_id].update({
                    "position_address": data.get("positionAddress", ""),
                    "base_amount": Decimal(str(data.get("baseTokenAmountAdded", 0))),
                    "quote_amount": Decimal(str(data.get("quoteTokenAmountAdded", 0))),
                    # SOL rent paid to create position
                    "position_rent": Decimal(str(data.get("positionRent", 0))),
                    # SOL transaction fee
                    "tx_fee": Decimal(str(data.get("fee", 0))),
                })
                return transaction_hash
            else:
                raise ValueError("No transaction hash returned from gateway")
        except asyncio.CancelledError:
            raise
        except Exception as e:
            self._handle_operation_failure(order_id, trading_pair, "opening CLMM position", e)
            raise  # Re-raise so executor can catch and retry if needed

    async def _amm_add_liquidity(
        self,
        trade_type: TradeType,
        order_id: str,
        trading_pair: str,
        price: float,
        base_token_amount: float,
        quote_token_amount: float,
        slippage_pct: Optional[float] = None,
    ):
        """
        Opens a regular AMM liquidity position.

        :param trade_type: The trade type (should be RANGE)
        :param order_id: Internal order id (also called client_order_id)
        :param trading_pair: The trading pair for the position
        :param price: The price for the position
        :param base_token_amount: Amount of base token to add (required)
        :param quote_token_amount: Amount of quote token to add (required)
        :param slippage_pct: Maximum allowed slippage percentage
        """
        # Check connector type is AMM
        if get_connector_type(self.connector_name) != ConnectorType.AMM:
            raise ValueError(f"Connector {self.connector_name} is not of type AMM.")

        # Split trading_pair to get base and quote tokens
        tokens = trading_pair.split("-")
        if len(tokens) != 2:
            raise ValueError(f"Invalid trading pair format: {trading_pair}")

        base_token, quote_token = tokens

        # Calculate the total amount in base token units
        quote_amount_in_base = quote_token_amount / price if price > 0 else 0.0
        total_amount_in_base = base_token_amount + quote_amount_in_base

        # Start tracking order with calculated amount
        self.start_tracking_order(order_id=order_id,
                                  trading_pair=trading_pair,
                                  trade_type=trade_type,
                                  price=Decimal(str(price)),
                                  amount=Decimal(str(total_amount_in_base)),
                                  order_type=OrderType.AMM_ADD)

        # Get pool address for the trading pair
        pool_address = await self.get_pool_address(trading_pair)
        if not pool_address:
            raise ValueError(f"Could not find pool for {trading_pair}")

        # Add liquidity to AMM pool
        try:
            transaction_result = await self._get_gateway_instance().amm_add_liquidity(
                connector=self.connector_name,
                network=self.network,
                wallet_address=self.address,
                pool_address=pool_address,
                base_token_amount=base_token_amount,
                quote_token_amount=quote_token_amount,
                slippage_pct=slippage_pct
            )
            transaction_hash: Optional[str] = transaction_result.get("signature")
            if transaction_hash is not None and transaction_hash != "":
                self.update_order_from_hash(order_id, trading_pair, transaction_hash, transaction_result)
                return transaction_hash
            else:
                raise ValueError("No transaction hash returned from gateway")
        except asyncio.CancelledError:
            raise
        except Exception as e:
            self._handle_operation_failure(order_id, trading_pair, "opening AMM position", e)

    def remove_liquidity(
        self,
        trading_pair: str,
        position_address: Optional[str] = None,
        percentage: float = 100.0,
        **request_args
    ) -> str:
        """
        Removes liquidity from a position - either concentrated (CLMM) or regular (AMM) based on the connector type.
        :param trading_pair: The market trading pair
        :param position_address: The address of the position (required for CLMM, optional for AMM)
        :param percentage: Percentage of liquidity to remove (defaults to 100%)
        :return: A newly created order id (internal).
        """
        connector_type = get_connector_type(self.connector_name)

        # Verify we have a position address for CLMM positions
        if connector_type == ConnectorType.CLMM and position_address is None:
            raise ValueError("position_address is required to close a CLMM position")

        trade_type: TradeType = TradeType.RANGE
        order_id: str = self.create_market_order_id(trade_type, trading_pair)

        # Call appropriate function based on connector type and percentage
        if connector_type == ConnectorType.CLMM:
            if percentage == 100.0:
                # Complete close for CLMM
                safe_ensure_future(self._clmm_close_position(trade_type, order_id, trading_pair, position_address, **request_args))
            else:
                # Partial removal for CLMM
                safe_ensure_future(self._clmm_remove_liquidity(trade_type, order_id, trading_pair, position_address, percentage, **request_args))
        elif connector_type == ConnectorType.AMM:
            # AMM always uses remove_liquidity
            safe_ensure_future(self._amm_remove_liquidity(trade_type, order_id, trading_pair, percentage, **request_args))
        else:
            raise ValueError(f"Connector type {connector_type} does not support liquidity provision")

        return order_id

    async def _clmm_close_position(
        self,
        trade_type: TradeType,
        order_id: str,
        trading_pair: str,
        position_address: str,
        fail_silently: bool = False,
    ):
        """
        Closes a concentrated liquidity position for the given position address.

        :param trade_type: The trade type (should be RANGE)
        :param order_id: Internal order id (also called client_order_id)
        :param trading_pair: The trading pair for the position
        :param position_address: The address of the position to close
        :param fail_silently: Whether to fail silently on error
        """
        # Check connector type is CLMM
        if get_connector_type(self.connector_name) != ConnectorType.CLMM:
            raise ValueError(f"Connector {self.connector_name} is not of type CLMM.")

        # Start tracking order
        self.start_tracking_order(order_id=order_id,
                                  trading_pair=trading_pair,
                                  trade_type=trade_type,
                                  order_type=OrderType.AMM_REMOVE)

        # Store metadata for event triggering (will be enriched with response data)
        self._lp_orders_metadata[order_id] = {
            "operation": "remove",
            "position_address": position_address,
        }

        try:
            transaction_result = await self._get_gateway_instance().clmm_close_position(
                connector=self.connector_name,
                network=self.network,
                wallet_address=self.address,
                position_address=position_address,
                fail_silently=fail_silently
            )
            transaction_hash: Optional[str] = transaction_result.get("signature")
            if transaction_hash is not None and transaction_hash != "":
                self.update_order_from_hash(order_id, trading_pair, transaction_hash, transaction_result)
                # Store response data in metadata for P&L tracking
                # Gateway returns positive values for token amounts
                data = transaction_result.get("data", {})
                self._lp_orders_metadata[order_id].update({
                    "base_amount": Decimal(str(data.get("baseTokenAmountRemoved", 0))),
                    "quote_amount": Decimal(str(data.get("quoteTokenAmountRemoved", 0))),
                    "base_fee": Decimal(str(data.get("baseFeeAmountCollected", 0))),
                    "quote_fee": Decimal(str(data.get("quoteFeeAmountCollected", 0))),
                    # SOL rent refunded on close
                    "position_rent_refunded": Decimal(str(data.get("positionRentRefunded", 0))),
                    # SOL transaction fee
                    "tx_fee": Decimal(str(data.get("fee", 0))),
                })
                return transaction_hash
            else:
                raise ValueError("No transaction hash returned from gateway")
        except asyncio.CancelledError:
            raise
        except Exception as e:
            self._handle_operation_failure(order_id, trading_pair, "closing CLMM position", e)
            raise  # Re-raise so executor can catch and retry if needed

    async def _clmm_remove_liquidity(
        self,
        trade_type: TradeType,
        order_id: str,
        trading_pair: str,
        position_address: str,
        percentage: float = 100.0,
        fail_silently: bool = False,
    ):
        """
        Removes liquidity from a CLMM position (partial removal).

        :param trade_type: The trade type (should be RANGE)
        :param order_id: Internal order id (also called client_order_id)
        :param trading_pair: The trading pair for the position
        :param position_address: The address of the position
        :param percentage: Percentage of liquidity to remove (0-100)
        :param fail_silently: Whether to fail silently on error
        """
        # Check connector type is CLMM
        if get_connector_type(self.connector_name) != ConnectorType.CLMM:
            raise ValueError(f"Connector {self.connector_name} is not of type CLMM.")

        # Start tracking order
        self.start_tracking_order(order_id=order_id,
                                  trading_pair=trading_pair,
                                  trade_type=trade_type,
                                  order_type=OrderType.AMM_REMOVE)

        # Store metadata for event triggering (will be enriched with response data)
        self._lp_orders_metadata[order_id] = {
            "operation": "remove",
            "position_address": position_address,
        }

        try:
            transaction_result = await self._get_gateway_instance().clmm_remove_liquidity(
                connector=self.connector_name,
                network=self.network,
                wallet_address=self.address,
                position_address=position_address,
                percentage=percentage,
                fail_silently=fail_silently
            )
            transaction_hash: Optional[str] = transaction_result.get("signature")
            if transaction_hash is not None and transaction_hash != "":
                self.update_order_from_hash(order_id, trading_pair, transaction_hash, transaction_result)
                # Store response data in metadata for P&L tracking
                # Gateway returns positive values for token amounts
                data = transaction_result.get("data", {})
                self._lp_orders_metadata[order_id].update({
                    "base_amount": Decimal(str(data.get("baseTokenAmountRemoved", 0))),
                    "quote_amount": Decimal(str(data.get("quoteTokenAmountRemoved", 0))),
                    "base_fee": Decimal(str(data.get("baseFeeAmountCollected", 0))),
                    "quote_fee": Decimal(str(data.get("quoteFeeAmountCollected", 0))),
                    # SOL rent refunded on close
                    "position_rent_refunded": Decimal(str(data.get("positionRentRefunded", 0))),
                    # SOL transaction fee
                    "tx_fee": Decimal(str(data.get("fee", 0))),
                })
                return transaction_hash
            else:
                raise ValueError("No transaction hash returned from gateway")
        except asyncio.CancelledError:
            raise
        except Exception as e:
            self._handle_operation_failure(order_id, trading_pair, "removing CLMM liquidity", e)

    async def _amm_remove_liquidity(
        self,
        trade_type: TradeType,
        order_id: str,
        trading_pair: str,
        percentage: float = 100.0,
        fail_silently: bool = False,
    ):
        """
        Closes an AMM liquidity position by removing specified percentage of liquidity.

        :param trade_type: The trade type (should be RANGE)
        :param order_id: Internal order id (also called client_order_id)
        :param trading_pair: The trading pair for the position
        :param percentage: Percentage of liquidity to remove (0-100)
        :param fail_silently: Whether to fail silently on error
        """
        # Check connector type is AMM
        if get_connector_type(self.connector_name) != ConnectorType.AMM:
            raise ValueError(f"Connector {self.connector_name} is not of type AMM.")

        # Get pool address for the trading pair
        pool_address = await self.get_pool_address(trading_pair)
        if not pool_address:
            raise ValueError(f"Could not find pool for {trading_pair}")

        # Start tracking order
        self.start_tracking_order(order_id=order_id,
                                  trading_pair=trading_pair,
                                  trade_type=trade_type,
                                  order_type=OrderType.AMM_REMOVE)

        try:
            transaction_result = await self._get_gateway_instance().amm_remove_liquidity(
                connector=self.connector_name,
                network=self.network,
                wallet_address=self.address,
                pool_address=pool_address,
                percentage=percentage,
                fail_silently=fail_silently
            )
            transaction_hash: Optional[str] = transaction_result.get("signature")
            if transaction_hash is not None and transaction_hash != "":
                self.update_order_from_hash(order_id, trading_pair, transaction_hash, transaction_result)
                return transaction_hash
            else:
                raise ValueError("No transaction hash returned from gateway")
        except asyncio.CancelledError:
            raise
        except Exception as e:
            self._handle_operation_failure(order_id, trading_pair, "closing AMM position", e)

    async def amm_add_liquidity(
        self,
        trading_pair: str,
        base_token_amount: float,
        quote_token_amount: float,
        slippage_pct: Optional[float] = None,
        fail_silently: bool = False
    ) -> Dict[str, Any]:
        """
        Adds liquidity to an AMM pool.
        :param trading_pair: The trading pair for the position
        :param base_token_amount: Amount of base token to add
        :param quote_token_amount: Amount of quote token to add
        :param slippage_pct: Maximum allowed slippage percentage
        :param fail_silently: Whether to fail silently on error
        :return: Response from the gateway API
        """
        # Check connector type is AMM
        if get_connector_type(self.connector_name) != ConnectorType.AMM:
            raise ValueError(f"Connector {self.connector_name} is not of type AMM.")

        # Get pool address for the trading pair
        pool_address = await self.get_pool_address(trading_pair)
        if not pool_address:
            raise ValueError(f"Could not find pool for {trading_pair}")

        order_id: str = self.create_market_order_id(TradeType.RANGE, trading_pair)
        self.start_tracking_order(order_id=order_id,
                                  trading_pair=trading_pair,
                                  trade_type=TradeType.RANGE)
        try:
            transaction_result = await self._get_gateway_instance().amm_add_liquidity(
                connector=self.connector_name,
                network=self.network,
                wallet_address=self.address,
                pool_address=pool_address,
                base_token_amount=base_token_amount,
                quote_token_amount=quote_token_amount,
                slippage_pct=slippage_pct,
                fail_silently=fail_silently
            )
            transaction_hash: Optional[str] = transaction_result.get("signature")
            if transaction_hash is not None and transaction_hash != "":
                self.update_order_from_hash(order_id, trading_pair, transaction_hash, transaction_result)
            else:
                raise ValueError
        except asyncio.CancelledError:
            raise
        except Exception as e:
            self._handle_operation_failure(order_id, trading_pair, "adding AMM liquidity", e)

    async def amm_remove_liquidity(
        self,
        trading_pair: str,
        percentage: float,
        fail_silently: bool = False
    ) -> Dict[str, Any]:
        """
        Removes liquidity from an AMM pool.
        :param trading_pair: The trading pair for the position
        :param percentage: Percentage of liquidity to remove (0-100)
        :param fail_silently: Whether to fail silently on error
        :return: Response from the gateway API
        """
        # Check connector type is AMM
        if get_connector_type(self.connector_name) != ConnectorType.AMM:
            raise ValueError(f"Connector {self.connector_name} is not of type AMM.")

        # Get pool address for the trading pair
        pool_address = await self.get_pool_address(trading_pair)
        if not pool_address:
            raise ValueError(f"Could not find pool for {trading_pair}")

        order_id: str = self.create_market_order_id(TradeType.RANGE, trading_pair)
        self.start_tracking_order(order_id=order_id,
                                  trading_pair=trading_pair,
                                  trade_type=TradeType.RANGE)
        try:
            transaction_result = await self._get_gateway_instance().amm_remove_liquidity(
                connector=self.connector_name,
                network=self.network,
                wallet_address=self.address,
                pool_address=pool_address,
                percentage=percentage,
                fail_silently=fail_silently
            )
            transaction_hash: Optional[str] = transaction_result.get("signature")
            if transaction_hash is not None and transaction_hash != "":
                self.update_order_from_hash(order_id, trading_pair, transaction_hash, transaction_result)
            else:
                raise ValueError
        except asyncio.CancelledError:
            raise
        except Exception as e:
            self._handle_operation_failure(order_id, trading_pair, "removing AMM liquidity", e)

    @async_ttl_cache(ttl=5, maxsize=10)
    async def get_position_info(
        self,
        trading_pair: str,
        position_address: Optional[str] = None
    ) -> Union[AMMPositionInfo, CLMMPositionInfo, None]:
        """
        Retrieves position information for a given liquidity position.

        :param trading_pair: The trading pair for the position
        :param position_address: The address of the position (required for CLMM, optional for AMM)
        :return: Position information from gateway, validated against appropriate schema
        """
        try:
            # Split trading_pair to get base and quote tokens
            tokens = trading_pair.split("-")
            if len(tokens) != 2:
                raise ValueError(f"Invalid trading pair format: {trading_pair}")

            base_token, quote_token = tokens

            connector_type = get_connector_type(self.connector_name)
            if connector_type == ConnectorType.CLMM:
                if position_address is None:
                    raise ValueError("position_address is required for CLMM positions")

                resp: Dict[str, Any] = await self._get_gateway_instance().clmm_position_info(
                    connector=self.connector_name,
                    network=self.network,
                    position_address=position_address,
                    wallet_address=self.address,
                )
                # Validate response against CLMM schema
                return CLMMPositionInfo(**resp) if resp else None

            elif connector_type == ConnectorType.AMM:
                resp: Dict[str, Any] = await self._get_gateway_instance().amm_position_info(
                    connector=self.connector_name,
                    network=self.network,
                    pool_address=position_address,
                    wallet_address=self.address,
                )
                # Validate response against AMM schema
                return AMMPositionInfo(**resp) if resp else None

            else:
                raise ValueError(f"Connector type {connector_type} does not support liquidity positions")

        except asyncio.CancelledError:
            raise
        except Exception as e:
            addr_info = f"position {position_address}" if position_address else trading_pair
            self.logger().network(
                f"Error fetching position info for {addr_info} on {self.connector_name}.",
                exc_info=True,
                app_warning_msg=str(e)
            )
            return None

    async def get_user_positions(self, pool_address: Optional[str] = None) -> List[Union[AMMPositionInfo, CLMMPositionInfo]]:
        """
        Fetch all user positions for this connector and wallet.

        :param pool_address: Optional pool address to filter positions (required for AMM)
        :return: List of position information objects
        """
        positions = []

        try:
            connector_type = get_connector_type(self.connector_name)

            if connector_type == ConnectorType.CLMM:
                # For CLMM, use positions-owned endpoint
                # Note: Gateway API doesn't support poolAddress filtering, so we filter client-side
                response = await self._get_gateway_instance().clmm_positions_owned(
                    connector=self.connector_name,
                    network=self.network,
                    wallet_address=self.address,
                    pool_address=None  # Gateway doesn't support this parameter
                )
            else:
                # For AMM, we need a pool address
                if not pool_address:
                    self.logger().warning("AMM position fetching requires a pool address")
                    return []

                # For AMM, get position info directly from the pool
                # We'll need to get pool info first to extract tokens
                pool_resp = await self._get_gateway_instance().pool_info(
                    connector=self.connector_name,
                    network=self.network,
                    pool_address=pool_address
                )

                if not pool_resp:
                    return []

                # Now get the position info
                resp = await self._get_gateway_instance().amm_position_info(
                    connector=self.connector_name,
                    network=self.network,
                    pool_address=pool_address,
                    wallet_address=self.address
                )

                if resp:
                    position = AMMPositionInfo(**resp)
                    # Get token symbols from loaded token data
                    base_token_info = self.get_token_by_address(position.base_token_address)
                    quote_token_info = self.get_token_by_address(position.quote_token_address)

                    # Use symbol if found, otherwise use address
                    position.base_token = base_token_info.get("symbol", position.base_token_address) if base_token_info else position.base_token_address
                    position.quote_token = quote_token_info.get("symbol", position.quote_token_address) if quote_token_info else position.quote_token_address
                    return [position]
                else:
                    return []

            # Parse position data based on connector type (for CLMM)
            # Handle case where response might be a list directly or a dict with 'positions' key
            positions_list = response if isinstance(response, list) else response.get("positions", [])
            for pos_data in positions_list:
                try:
                    if connector_type == ConnectorType.CLMM:
                        position = CLMMPositionInfo(**pos_data)

                        # Get token symbols from loaded token data
                        base_token_info = self.get_token_by_address(position.base_token_address)
                        quote_token_info = self.get_token_by_address(position.quote_token_address)

                        # Use symbol if found, otherwise use address
                        position.base_token = base_token_info.get("symbol", position.base_token_address) if base_token_info else position.base_token_address
                        position.quote_token = quote_token_info.get("symbol", position.quote_token_address) if quote_token_info else position.quote_token_address

                        positions.append(position)
                    else:
                        position = AMMPositionInfo(**pos_data)

                        # Get token symbols from loaded token data
                        base_token_info = self.get_token_by_address(position.base_token_address)
                        quote_token_info = self.get_token_by_address(position.quote_token_address)

                        # Use symbol if found, otherwise use address
                        position.base_token = base_token_info.get("symbol", position.base_token_address) if base_token_info else position.base_token_address
                        position.quote_token = quote_token_info.get("symbol", position.quote_token_address) if quote_token_info else position.quote_token_address

                        positions.append(position)

                except Exception as e:
                    self.logger().error(f"Error parsing position data: {e}", exc_info=True)
                    continue

            # Filter positions by pool_address if specified (client-side filtering)
            if pool_address and connector_type == ConnectorType.CLMM:
                positions = [p for p in positions if hasattr(p, 'pool_address') and p.pool_address == pool_address]

        except Exception as e:
            self.logger().error(f"Error fetching positions: {e}", exc_info=True)

        return positions



import inspect

from sqlalchemy import JSON, Column, Index, Text

from hummingbot.model import HummingbotBase
from hummingbot.model.decimal_type_decorator import SqliteDecimal


class MarketData(HummingbotBase):
    __tablename__ = "MarketData"
    __table_args__ = (
        Index("timestamp", "exchange", "trading_pair"),
    )

    timestamp = Column(SqliteDecimal(6), primary_key=True, nullable=False)
    exchange = Column(Text, nullable=False)
    trading_pair = Column(Text, nullable=False)
    mid_price = Column(SqliteDecimal(6), nullable=False)
    best_bid = Column(SqliteDecimal(6), nullable=False)
    best_ask = Column(SqliteDecimal(6), nullable=False)
    order_book = Column(JSON)

    def __repr__(self) -> str:
        list_of_fields = [f"{name}: {value}" for name, value in inspect.getmembers(self) if isinstance(value, Column)]
        return ','.join(list_of_fields)



# distutils: language=c++

from libcpp.string cimport string

cdef extern from "../cpp/OrderExpirationEntry.h":
    cdef cppclass OrderExpirationEntry:
        OrderExpirationEntry()
        OrderExpirationEntry(string trading_pair,
                             string order_id,
                             double timestamp,
                             double expiration)
        OrderExpirationEntry(const OrderExpirationEntry &other)
        OrderExpirationEntry &operator=(const OrderExpirationEntry &other)
        string getClientOrderID()
        string getTradingPair()
        double getTimestamp()
        double getExpiration()
        double getExpirationTimestamp()



import asyncio
import time
from decimal import Decimal
from typing import Any, Dict, List, Optional, Tuple

from bidict import bidict

import hummingbot.connector.derivative.dydx_v4_perpetual.dydx_v4_perpetual_constants as CONSTANTS
from hummingbot.connector.constants import s_decimal_0, s_decimal_NaN
from hummingbot.connector.derivative.dydx_v4_perpetual import dydx_v4_perpetual_web_utils as web_utils
from hummingbot.connector.derivative.dydx_v4_perpetual.data_sources.dydx_v4_data_source import DydxPerpetualV4Client
from hummingbot.connector.derivative.dydx_v4_perpetual.dydx_v4_perpetual_api_order_book_data_source import (
    DydxV4PerpetualAPIOrderBookDataSource,
)
from hummingbot.connector.derivative.dydx_v4_perpetual.dydx_v4_perpetual_user_stream_data_source import (
    DydxV4PerpetualUserStreamDataSource,
)
from hummingbot.connector.derivative.position import Position
from hummingbot.connector.perpetual_derivative_py_base import PerpetualDerivativePyBase
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.connector.utils import combine_to_hb_trading_pair, get_new_numeric_client_order_id
from hummingbot.core.api_throttler.data_types import RateLimit
from hummingbot.core.data_type.common import OrderType, PositionAction, PositionMode, PositionSide, TradeType
from hummingbot.core.data_type.in_flight_order import InFlightOrder, OrderState, OrderUpdate, TradeUpdate
from hummingbot.core.data_type.trade_fee import TokenAmount, TradeFeeBase
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.event.events import AccountEvent, PositionModeChangeEvent
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.core.utils.estimate_fee import build_perpetual_trade_fee
from hummingbot.core.utils.tracking_nonce import NonceCreator
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory


class DydxV4PerpetualDerivative(PerpetualDerivativePyBase):
    web_utils = web_utils

    def __init__(
            self,
            dydx_v4_perpetual_secret_phrase: str,
            dydx_v4_perpetual_chain_address: str,
            balance_asset_limit: Optional[Dict[str, Dict[str, Decimal]]] = None,
            rate_limits_share_pct: Decimal = Decimal("100"),
            trading_pairs: Optional[List[str]] = None,
            trading_required: bool = True,
            domain: str = CONSTANTS.DEFAULT_DOMAIN,
    ):
        self._dydx_v4_perpetual_secret_phrase = dydx_v4_perpetual_secret_phrase
        self._dydx_v4_perpetual_chain_address = dydx_v4_perpetual_chain_address
        self._trading_pairs = trading_pairs
        self._trading_required = trading_required
        self._domain = domain
        self._client_order_id_nonce_provider = NonceCreator.for_microseconds()

        self._tx_client: DydxPerpetualV4Client = self._create_tx_client()

        self._margin_fractions = {}
        self._position_id = None

        self._allocated_collateral = {}
        self.subaccount_id = 0

        super().__init__(balance_asset_limit, rate_limits_share_pct)

    @property
    def name(self) -> str:
        return CONSTANTS.EXCHANGE_NAME

    @property
    def authenticator(self) -> AuthBase:
        return None

    @property
    def rate_limits_rules(self) -> List[RateLimit]:
        return CONSTANTS.RATE_LIMITS

    @property
    def domain(self) -> str:
        return self._domain

    @property
    def client_order_id_max_length(self) -> int:
        return CONSTANTS.MAX_ID_LEN

    @property
    def client_order_id_prefix(self) -> str:
        return CONSTANTS.HBOT_BROKER_ID

    @property
    def trading_rules_request_path(self) -> str:
        return CONSTANTS.PATH_MARKETS

    @property
    def trading_pairs_request_path(self) -> str:
        return CONSTANTS.PATH_MARKETS

    @property
    def check_network_request_path(self) -> str:
        return CONSTANTS.PATH_TIME

    @property
    def trading_pairs(self) -> List[str]:
        return self._trading_pairs

    @property
    def is_cancel_request_in_exchange_synchronous(self) -> bool:
        return True

    @property
    def is_trading_required(self) -> bool:
        return self._trading_required

    @property
    def funding_fee_poll_interval(self) -> int:
        return 120

    def supported_order_types(self) -> List[OrderType]:
        return [OrderType.LIMIT, OrderType.LIMIT_MAKER, OrderType.MARKET]

    def _is_request_exception_related_to_time_synchronizer(self, request_exception: Exception) -> bool:
        return False

    def _is_request_result_an_error_related_to_time_synchronizer(self, request_result: Dict[str, Any]) -> bool:
        if "errors" in request_result and "msg" in request_result["errors"]:
            if "Timestamp must be within" in request_result["errors"]["msg"]:
                return True
        return False

    async def _make_trading_rules_request(self) -> Any:
        exchange_info = await self._api_get(path_url=self.trading_rules_request_path,
                                            params={})
        return exchange_info

    async def _make_trading_pairs_request(self) -> Any:
        exchange_info = await self._api_get(path_url=self.trading_pairs_request_path,
                                            params={})
        return exchange_info

    def _is_order_not_found_during_status_update_error(self, status_update_exception: Exception) -> bool:
        # TODO: implement this method correctly for the connector
        # The default implementation was added when the functionality to detect not found orders was introduced in the
        # ExchangePyBase class. Also fix the unit test test_lost_order_removed_if_not_found_during_order_status_update
        # when replacing the dummy implementation
        return False

    def _is_order_not_found_during_cancelation_error(self, cancelation_exception: Exception) -> bool:
        # The default implementation was added when the functionality to detect not found orders was introduced in the
        # ExchangePyBase class. Also fix the unit test test_cancel_order_not_found_in_the_exchange when replacing the
        # dummy implementation
        return False

    async def start_network(self):
        await super().start_network()
        await self._update_trading_rules()
        await self._tx_client.initialize_trading_account()

    async def _place_cancel(self, order_id: str, tracked_order: InFlightOrder):
        async with self._throttler.execute_task(limit_id=CONSTANTS.LIMIT_ID_ORDER_CANCEL):
            if not self._margin_fractions:
                await self._update_trading_rules()
            for i in range(3):
                resp = await self._tx_client.cancel_order(
                    client_id=int(tracked_order.client_order_id),
                    clob_pair_id=self._margin_fractions[tracked_order.trading_pair]["clob_pair_id"],
                    order_flags=CONSTANTS.ORDER_FLAGS_LONG_TERM,
                    good_til_block_time=int(time.time()) + CONSTANTS.ORDER_EXPIRATION
                )
                if CONSTANTS.ACCOUNT_SEQUENCE_MISMATCH_ERROR in resp['raw_log']:
                    self.logger().warning(
                        f"Failed to cancel order {tracked_order.client_order_id} (retry {i + 1}), {resp['raw_log']}")
                    await asyncio.sleep(1)
                    continue
                else:
                    break
        if resp["raw_log"] != "[]" and CONSTANTS.ERR_MSG_NO_ORDER_FOUND not in resp['raw_log']:
            raise ValueError(f"Error sending the order cancel transaction ({resp['raw_log']})")
        else:
            return True

    def buy(self,
            trading_pair: str,
            amount: Decimal,
            order_type=OrderType.LIMIT,
            price: Decimal = s_decimal_NaN,
            **kwargs) -> str:
        """
        Creates a promise to create a buy order using the parameters

        :param trading_pair: the token pair to operate with
        :param amount: the order amount
        :param order_type: the type of order to create (MARKET, LIMIT, LIMIT_MAKER)
        :param price: the order price

        :return: the id assigned by the connector to the order (the client id)
        """
        order_id = str(get_new_numeric_client_order_id(
            nonce_creator=self._client_order_id_nonce_provider,
            max_id_bit_count=CONSTANTS.MAX_ID_BIT_COUNT,
        ))
        safe_ensure_future(self._create_order(
            trade_type=TradeType.BUY,
            order_id=order_id,
            trading_pair=trading_pair,
            amount=amount,
            order_type=order_type,
            price=price,
            **kwargs))
        return order_id

    def sell(self,
             trading_pair: str,
             amount: Decimal,
             order_type: OrderType = OrderType.LIMIT,
             price: Decimal = s_decimal_NaN,
             **kwargs) -> str:
        """
        Creates a promise to create a sell order using the parameters.
        :param trading_pair: the token pair to operate with
        :param amount: the order amount
        :param order_type: the type of order to create (MARKET, LIMIT, LIMIT_MAKER)
        :param price: the order price
        :return: the id assigned by the connector to the order (the client id)
        """
        order_id = str(get_new_numeric_client_order_id(
            nonce_creator=self._client_order_id_nonce_provider,
            max_id_bit_count=CONSTANTS.MAX_ID_BIT_COUNT,
        ))
        safe_ensure_future(self._create_order(
            trade_type=TradeType.SELL,
            order_id=order_id,
            trading_pair=trading_pair,
            amount=amount,
            order_type=order_type,
            price=price,
            **kwargs))
        return order_id

    async def _place_order(
            self,
            order_id: str,
            trading_pair: str,
            amount: Decimal,
            trade_type: TradeType,
            order_type: OrderType,
            price: Decimal,
            position_action: PositionAction = PositionAction.NIL,
            **kwargs,
    ):
        if not self._margin_fractions:
            await self._update_trading_rules()
        if order_type.is_limit_type():
            _order_type = "LIMIT"
            limit_id = CONSTANTS.LIMIT_LONG_TERM_ORDER_PLACE
        else:
            limit_id = CONSTANTS.MARKET_SHORT_TERM_ORDER_PLACE
            _order_type = "MARKET"
            if trade_type.name.lower() == 'buy':
                # The price needs to be relatively high before the transaction, whether the test will be cancelled
                price = Decimal("1.5") * self.get_price_for_volume(
                    trading_pair,
                    True,
                    amount
                ).result_price
            else:
                price = Decimal("0.75") * self.get_price_for_volume(
                    trading_pair,
                    False,
                    amount
                ).result_price
            price = self.quantize_order_price(trading_pair, price)
        side = "BUY" if trade_type == TradeType.BUY else "SELL"
        expiration = CONSTANTS.ORDER_EXPIRATION
        reduce_only = False

        post_only = order_type is OrderType.LIMIT_MAKER
        market = await self.exchange_symbol_associated_to_pair(trading_pair)
        try:
            async with self._throttler.execute_task(limit_id=limit_id):
                for i in range(3):
                    resp = await self._tx_client.place_order(
                        market=market,
                        type=_order_type,
                        side=side,
                        price=price,
                        size=amount,
                        client_id=int(order_id),
                        post_only=post_only,
                        reduce_only=reduce_only,
                        good_til_time_in_seconds=expiration,
                    )
                    if CONSTANTS.ACCOUNT_SEQUENCE_MISMATCH_ERROR in resp['raw_log']:
                        self.logger().warning(f"Failed to submit order {order_id} (retry {i + 1}), {resp['raw_log']}")
                        await asyncio.sleep(1)
                        continue
                    else:
                        break
            if resp["raw_log"] != "[]" or resp["txhash"] in [None, ""]:
                raise ValueError(f"Error sending the order creation transaction ({resp['raw_log']})")
        except asyncio.CancelledError:
            raise
        except Exception:
            raise
        # no exchange order id
        return None, self._time()

    async def _place_order_and_process_update(self, order: InFlightOrder, **kwargs) -> str:
        exchange_order_id, update_timestamp = await self._place_order(
            order_id=order.client_order_id,
            trading_pair=order.trading_pair,
            amount=order.amount,
            trade_type=order.trade_type,
            order_type=order.order_type,
            price=order.price,
            **kwargs,
        )

        order_update: OrderUpdate = OrderUpdate(
            client_order_id=order.client_order_id,
            exchange_order_id=exchange_order_id,
            trading_pair=order.trading_pair,
            update_timestamp=update_timestamp,
            new_state=OrderState.OPEN,
        )
        self._order_tracker.process_order_update(order_update)

        return exchange_order_id

    def _on_order_failure(
            self,
            order_id: str,
            trading_pair: str,
            amount: Decimal,
            trade_type: TradeType,
            order_type: OrderType,
            price: Optional[Decimal],
            exception: Exception,
            **kwargs,
    ):
        self.logger().network(
            f"Error submitting {trade_type.name.lower()} {order_type.name.upper()} order to {self.name_cap} for "
            f"{amount} {trading_pair} {price}.",
            exc_info=exception,
            app_warning_msg=f"Failed to submit {trade_type.name.upper()} order to {self.name_cap}. Check API key and network connection."
        )
        self._update_order_after_failure(order_id=order_id, trading_pair=trading_pair)

    def _get_fee(
            self,
            base_currency: str,
            quote_currency: str,
            order_type: OrderType,
            order_side: TradeType,
            position_action: PositionAction,
            amount: Decimal,
            price: Decimal = s_decimal_NaN,
            is_maker: Optional[bool] = None,
    ) -> TradeFeeBase:
        is_maker = is_maker or False
        fee = build_perpetual_trade_fee(
            self.name,
            is_maker,
            position_action=position_action,
            base_currency=base_currency,
            quote_currency=quote_currency,
            order_type=order_type,
            order_side=order_side,
            amount=amount,
            price=price,
        )
        return fee

    async def _update_trading_fees(self):
        pass

    async def _user_stream_event_listener(self):

        async for event_message in self._iter_user_event_queue():
            try:
                event: Dict[str, Any] = event_message
                data: Dict[str, Any] = event["contents"]
                quote = "USD"
                if "subaccount" in data.keys() and len(data["subaccount"]) > 0:
                    self._account_balances[quote] = Decimal(data["subaccount"]["equity"])
                    self._account_available_balances[quote] = Decimal(data["subaccount"]["freeCollateral"])
                    if "openPerpetualPositions" in data["subaccount"]:
                        await self._process_open_positions(data["subaccount"]["openPerpetualPositions"])
                if "orders" in data.keys() and len(data["orders"]) > 0:
                    for order in data["orders"]:
                        client_order_id: str = order["clientId"]
                        exchange_order_id: str = order["id"]
                        tracked_order = self._order_tracker.all_updatable_orders.get(client_order_id)
                        trading_pair = await self.trading_pair_associated_to_exchange_symbol(order["ticker"])
                        if tracked_order is not None:
                            state = CONSTANTS.ORDER_STATE[order["status"]]
                            new_order_update: OrderUpdate = OrderUpdate(
                                trading_pair=tracked_order.trading_pair,
                                update_timestamp=self.current_timestamp,
                                new_state=state,
                                client_order_id=tracked_order.client_order_id,
                                exchange_order_id=exchange_order_id,
                            )
                            self._order_tracker.process_order_update(new_order_update)
                        # Processing all orders of the account, not just the client's
                        if order["status"] in ["OPEN"]:
                            initial_margin_requirement = (
                                Decimal(order["price"])
                                * Decimal(order["size"])
                                * self._margin_fractions[trading_pair]["initial"]
                            )
                            initial_margin_requirement = abs(initial_margin_requirement)
                            self._allocated_collateral[order["id"]] = initial_margin_requirement
                            # self._allocated_collateral_sum += initial_margin_requirement
                            self._account_available_balances[quote] -= initial_margin_requirement
                        if order["status"] in ["FILLED", "CANCELED"]:
                            if order["id"] in self._allocated_collateral:
                                # Only deduct orders that were previously in the OPEN state
                                # Some orders are filled instantly and reach only the PENDING state
                                # self._allocated_collateral_sum -= self._allocated_collateral[order["id"]]
                                self._account_available_balances[quote] += self._allocated_collateral[order["id"]]
                                del self._allocated_collateral[order["id"]]
                if "fills" in data.keys() and len(data["fills"]) > 0:
                    trade_updates = await self._process_ws_fills(data["fills"])
                    for trade_update in trade_updates:
                        self._order_tracker.process_trade_update(trade_update)

                if "perpetualPositions" in data.keys() and len(data["perpetualPositions"]) > 0:
                    # this is hit when a position is closed
                    positions = data["perpetualPositions"]
                    for position in positions:
                        trading_pair = position["market"]
                        if trading_pair not in self._trading_pairs:
                            continue
                        position_side = PositionSide[position["side"]]
                        pos_key = self._perpetual_trading.position_key(trading_pair, position_side)
                        amount = Decimal(position.get("size"))

                        if amount != s_decimal_0:
                            position = Position(
                                trading_pair=trading_pair,
                                position_side=position_side,
                                unrealized_pnl=Decimal(position.get("unrealizedPnl", 0)),
                                entry_price=Decimal(position.get("entryPrice")),
                                amount=amount,
                                leverage=self.get_leverage(trading_pair),
                            )
                            self._perpetual_trading.set_position(pos_key, position)
                        else:
                            if self._perpetual_trading.get_position(pos_key):
                                self._perpetual_trading.remove_position(pos_key)

                # if "fundingPayments" in data:
                #     if event["type"] != CONSTANTS.WS_TYPE_SUBSCRIBED:  # Only subsequent funding payments
                #         funding_payments = await self._process_funding_payments(data["fundingPayments"])
                #         for trading_pair in funding_payments:
                #             timestamp = funding_payments[trading_pair]["timestamp"]
                #             funding_rate = funding_payments[trading_pair]["funding_rate"]
                #             payment = funding_payments[trading_pair]["payment"]
                #             self._emit_funding_payment_event(trading_pair, timestamp, funding_rate, payment, False)

            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().error("Unexpected error in user stream listener loop.", exc_info=True)

    async def _format_trading_rules(self, exchange_info_dict: Dict[str, Any]) -> List[TradingRule]:
        trading_rules = []
        markets_info = exchange_info_dict["markets"]
        for market_name, market_info in markets_info.items():
            if web_utils.is_exchange_information_valid(market_info):
                trading_pair = await self.trading_pair_associated_to_exchange_symbol(symbol=market_name)
                market = markets_info[market_name]
                try:
                    collateral_token = CONSTANTS.CURRENCY
                    trading_rules += [
                        TradingRule(
                            trading_pair=trading_pair,
                            min_price_increment=Decimal(market["tickSize"]),
                            min_base_amount_increment=Decimal(market["stepSize"]),
                            supports_limit_orders=True,
                            supports_market_orders=True,
                            buy_order_collateral_token=collateral_token,
                            sell_order_collateral_token=collateral_token,
                        )
                    ]
                    self._margin_fractions[trading_pair] = {
                        "initial": Decimal(market["initialMarginFraction"]),
                        "maintenance": Decimal(market["maintenanceMarginFraction"]),
                        "clob_pair_id": market["clobPairId"],
                        "atomicResolution": market["atomicResolution"],
                        "stepBaseQuantums": market["stepBaseQuantums"],
                        "quantumConversionExponent": market["quantumConversionExponent"],
                        "subticksPerTick": market["subticksPerTick"],
                    }
                except Exception:
                    self.logger().exception("Error updating trading rules")
        return trading_rules

    async def _update_balances(self):
        path = f"{CONSTANTS.PATH_SUBACCOUNT}/{self._dydx_v4_perpetual_chain_address}/subaccountNumber/{self.subaccount_id}"
        response: Dict[str, Dict[str, Any]] = await self._api_get(
            path_url=path, params={}, limit_id=CONSTANTS.PATH_SUBACCOUNT
        )
        quote = CONSTANTS.CURRENCY
        self._account_available_balances.clear()
        self._account_balances.clear()

        self._account_balances[quote] = Decimal(response["subaccount"]["equity"])
        self._account_available_balances[quote] = Decimal(response["subaccount"]["freeCollateral"])

    async def _process_ws_fills(self, fills_data: List) -> List[TradeUpdate]:
        trade_updates = []

        for fill_data in fills_data:
            exchange_order_id: str = fill_data["orderId"]
            all_orders = self._order_tracker.all_fillable_orders
            try:
                for k, v in all_orders.items():
                    await v.get_exchange_order_id()
            except Exception as e:
                self.logger().info(
                    f"Unable to locate order {exchange_order_id} on exchange. Pending update from blockchain {e}")
            _cli_tracked_orders = [o for o in all_orders.values() if exchange_order_id == o.exchange_order_id]
            if len(_cli_tracked_orders) == 0 or _cli_tracked_orders[0] is None:
                order_update: OrderUpdate = await self._request_order_status(tracked_order=None,
                                                                             exchange_order_id=exchange_order_id)
                # NOTE: Untracked order
                if order_update is None:
                    self.logger().debug(f"Received untracked order with exchange order id of {exchange_order_id}")
                    return trade_updates
                client_order_id = order_update.client_order_id
                tracked_order = self._order_tracker.all_updatable_orders.get(client_order_id)
            else:
                tracked_order = _cli_tracked_orders[0]
            trade_update = self._process_order_fills(fill_data=fill_data, order=tracked_order)
            if trade_update is not None:
                trade_updates.append(trade_update)
        return trade_updates

    async def _process_open_positions(self, open_positions: Dict):
        for market, position in open_positions.items():
            trading_pair = await self.trading_pair_associated_to_exchange_symbol(symbol=market)
            position_side = PositionSide[position["side"]]
            pos_key = self._perpetual_trading.position_key(trading_pair, position_side)
            amount = Decimal(position.get("size"))
            if amount != s_decimal_0:
                entry_price = Decimal(position.get("entryPrice"))
                unrealized_pnl = Decimal(position.get("unrealizedPnl"))
                position = Position(
                    trading_pair=trading_pair,
                    position_side=position_side,
                    unrealized_pnl=unrealized_pnl,
                    entry_price=entry_price,
                    amount=amount,
                    leverage=self.get_leverage(trading_pair),
                )
                self._perpetual_trading.set_position(pos_key, position)
            else:
                self._perpetual_trading.remove_position(pos_key)

    async def _all_trade_updates_for_order(self, order: InFlightOrder) -> List[TradeUpdate]:
        trade_updates = []

        if order.exchange_order_id is not None:
            try:
                all_fills_response = await self._request_order_fills(order=order)
                fills_data = all_fills_response["fills"]
                trade_updates_tmp = self._process_rest_fills(fills_data)
                trade_updates += trade_updates_tmp

            except IOError as ex:
                if not self._is_request_exception_related_to_time_synchronizer(request_exception=ex):
                    raise
        return trade_updates

    def _process_rest_fills(self, fills_data: List) -> List[TradeUpdate]:
        trade_updates = []
        all_fillable_orders_by_exchange_order_id = {
            order.exchange_order_id: order for order in self._order_tracker.all_fillable_orders.values()
        }
        for fill_data in fills_data:
            exchange_order_id: str = fill_data["orderId"]
            order = all_fillable_orders_by_exchange_order_id.get(exchange_order_id)
            trade_update = self._process_order_fills(fill_data=fill_data, order=order)
            if trade_update is not None:
                trade_updates.append(trade_update)
        return trade_updates

    def _process_order_fills(self, fill_data: Dict, order: InFlightOrder) -> Optional[TradeUpdate]:
        trade_update = None
        if order is not None:
            fee_asset = order.quote_asset
            fee_amount = Decimal(fill_data["fee"])
            flat_fees = [] if fee_amount == Decimal("0") else [TokenAmount(amount=fee_amount, token=fee_asset)]
            position_side = fill_data["side"]

            position_action = (PositionAction.OPEN
                               if (order.trade_type is TradeType.BUY and position_side == "BUY"
                                   or order.trade_type is TradeType.SELL and position_side == "SELL")
                               else PositionAction.CLOSE)

            fee = TradeFeeBase.new_perpetual_fee(
                fee_schema=self.trade_fee_schema(),
                position_action=position_action,
                percent_token=fee_asset,
                flat_fees=flat_fees,
            )

            trade_update = TradeUpdate(
                trade_id=fill_data["id"],
                client_order_id=order.client_order_id,
                exchange_order_id=order.exchange_order_id,
                trading_pair=order.trading_pair,
                fill_timestamp=fill_data["createdAt"],
                fill_price=Decimal(fill_data["price"]),
                fill_base_amount=Decimal(fill_data["size"]),
                fill_quote_amount=Decimal(fill_data["price"]) * Decimal(fill_data["size"]),
                fee=fee,
            )
        return trade_update

    async def _request_order_fills(self, order: InFlightOrder) -> Dict[str, Any]:

        body_params = {
            'address': self._dydx_v4_perpetual_chain_address,
            'subaccountNumber': self.subaccount_id,
            'marketType': 'PERPETUAL',
            'market': order.trading_pair,
            'limit': CONSTANTS.LAST_FILLS_MAX,
        }

        res = await self._api_get(
            path_url=CONSTANTS.PATH_FILLS,
            params=body_params,
        )
        return res

    async def _request_order_status(self, tracked_order: InFlightOrder, exchange_order_id=None) -> OrderUpdate:
        try:
            orders_rsp = await self._api_get(
                path_url=CONSTANTS.PATH_ORDERS,
                limit_id=CONSTANTS.PATH_ORDERS,
                params={
                    'address': self._dydx_v4_perpetual_chain_address,
                    'subaccountNumber': self.subaccount_id,
                    'goodTilBlockBeforeOrAt': CONSTANTS.TX_MAX_HEIGHT,
                    'limit': CONSTANTS.LAST_FILLS_MAX,
                }
            )
            if exchange_order_id:
                updated_order_data = next(
                    (order for order in orders_rsp if
                     order["id"] == exchange_order_id), None
                )
                if updated_order_data is None:
                    return None
                tracked_order = self._order_tracker.all_updatable_orders.get(str(updated_order_data["clientId"]))
            else:
                updated_order_data = next(
                    (order for order in orders_rsp if
                     int(order["clientId"]) == int(tracked_order.client_order_id)), None
                )

            if updated_order_data is None:
                # If the order is not found in the response, return an OrderUpdate with the same status as before
                return OrderUpdate(
                    client_order_id=tracked_order.client_order_id,
                    exchange_order_id=tracked_order.exchange_order_id,
                    trading_pair=tracked_order.trading_pair,
                    update_timestamp=self._time_synchronizer.time(),
                    new_state=tracked_order.current_state,
                )
            client_order_id = str(updated_order_data["clientId"])

            order_update: OrderUpdate = OrderUpdate(
                trading_pair=tracked_order.trading_pair,
                update_timestamp=self.current_timestamp,
                new_state=CONSTANTS.ORDER_STATE[updated_order_data["status"]],
                client_order_id=client_order_id,
                exchange_order_id=updated_order_data["id"],
            )
        except IOError as ex:
            if self._is_request_exception_related_to_time_synchronizer(request_exception=ex):
                order_update = OrderUpdate(
                    client_order_id=tracked_order.client_order_id,
                    trading_pair=tracked_order.trading_pair,
                    update_timestamp=self.current_timestamp,
                    new_state=tracked_order.current_state,
                )
            else:
                raise
        return order_update

    def _create_web_assistants_factory(self) -> WebAssistantsFactory:
        return web_utils.build_api_factory(
            throttler=self._throttler,
        )

    def _create_tx_client(self) -> DydxPerpetualV4Client:
        return DydxPerpetualV4Client(
            self._dydx_v4_perpetual_secret_phrase,
            self._dydx_v4_perpetual_chain_address,
            connector=self
        )

    def _create_order_book_data_source(self) -> DydxV4PerpetualAPIOrderBookDataSource:
        return DydxV4PerpetualAPIOrderBookDataSource(
            self.trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
            domain=self._domain,
        )

    def _create_user_stream_data_source(self) -> UserStreamTrackerDataSource:
        return DydxV4PerpetualUserStreamDataSource(api_factory=self._web_assistants_factory,
                                                   connector=self)

    def _initialize_trading_pair_symbols_from_exchange_info(self, exchange_info: Dict[str, Any]):
        markets = exchange_info["markets"]

        mapping = bidict()
        for key, val in markets.items():
            if web_utils.is_exchange_information_valid(val):
                exchange_symbol = val["ticker"]
                base = exchange_symbol.split("-")[0]
                quote = CONSTANTS.CURRENCY
                trading_pair = combine_to_hb_trading_pair(base, quote)
                if trading_pair in mapping.inverse:
                    self._resolve_trading_pair_symbols_duplicate(mapping, exchange_symbol, base, quote)
                else:
                    mapping[exchange_symbol] = trading_pair

        self._set_trading_pair_symbol_map(mapping)

    def _resolve_trading_pair_symbols_duplicate(self, mapping: bidict, new_exchange_symbol: str, base: str, quote: str):
        """Resolves name conflicts provoked by futures contracts.
        If the expected BASEQUOTE combination matches one of the exchange symbols, it is the one taken, otherwise,
        the trading pair is removed from the map and an error is logged.
        """
        expected_exchange_symbol = f"{base}{quote}"
        trading_pair = combine_to_hb_trading_pair(base, quote)
        current_exchange_symbol = mapping.inverse[trading_pair]
        if current_exchange_symbol == expected_exchange_symbol:
            pass
        elif new_exchange_symbol == expected_exchange_symbol:
            mapping.pop(current_exchange_symbol)
            mapping[new_exchange_symbol] = trading_pair
        else:
            self.logger().error(
                f"Could not resolve the exchange symbols {new_exchange_symbol} and {current_exchange_symbol}"
            )
            mapping.pop(current_exchange_symbol)

    async def _get_last_traded_price(self, trading_pair: str) -> float:
        exchange_symbol = await self.exchange_symbol_associated_to_pair(trading_pair)
        params = {}

        response: Dict[str, Dict[str, Any]] = await self._api_get(
            path_url=CONSTANTS.PATH_MARKETS, params=params, is_auth_required=False
        )
        price = float(response["markets"][exchange_symbol]["oraclePrice"])
        return price

    def supported_position_modes(self) -> List[PositionMode]:
        return [PositionMode.ONEWAY]

    def get_buy_collateral_token(self, trading_pair: str) -> str:
        trading_rule: TradingRule = self._trading_rules[trading_pair]
        return trading_rule.buy_order_collateral_token

    def get_sell_collateral_token(self, trading_pair: str) -> str:
        trading_rule: TradingRule = self._trading_rules[trading_pair]
        return trading_rule.sell_order_collateral_token

    async def _update_positions(self):
        params = {}
        path = f"{CONSTANTS.PATH_SUBACCOUNT}/{self._dydx_v4_perpetual_chain_address}/subaccountNumber/{self.subaccount_id}"
        response: Dict[str, Dict[str, Any]] = await self._api_get(
            path_url=path, params=params, limit_id=CONSTANTS.PATH_SUBACCOUNT
        )

        # account = await self._get_account()
        await self._process_open_positions(response["subaccount"]["openPerpetualPositions"])

    async def _trading_pair_position_mode_set(self, mode: PositionMode, trading_pair: str) -> Tuple[bool, str]:
        """
        :return: A tuple of boolean (true if success) and error message if the exchange returns one on failure.
        """
        if mode != PositionMode.ONEWAY:
            self.trigger_event(
                AccountEvent.PositionModeChangeFailed,
                PositionModeChangeEvent(
                    self.current_timestamp, trading_pair, mode, "dydx_v4 only supports the ONEWAY position mode."
                ),
            )
            self.logger().debug(
                f"dydx_v4 encountered a problem switching position mode to "
                f"{mode} for {trading_pair}"
                f" (dydx_v4 only supports the ONEWAY position mode)"
            )
        else:
            self._position_mode = PositionMode.ONEWAY
            super().set_position_mode(PositionMode.ONEWAY)
            self.trigger_event(
                AccountEvent.PositionModeChangeSucceeded,
                PositionModeChangeEvent(self.current_timestamp, trading_pair, mode),
            )
            self.logger().debug(f"dydx_v4 switching position mode to " f"{mode} for {trading_pair} succeeded.")

    async def _set_trading_pair_leverage(self, trading_pair: str, leverage: int) -> Tuple[bool, str]:
        success = True
        msg = ""

        response: Dict[str, Dict[str, Any]] = await self._api_get(
            path_url=CONSTANTS.PATH_MARKETS,
            is_auth_required=False,
        )

        if "markets" not in response:
            msg = "Failed to obtain markets information."
            success = False

        if success:
            markets_info = response["markets"][trading_pair]
            self._margin_fractions[trading_pair] = {
                "initial": Decimal(markets_info["initialMarginFraction"]),
                "maintenance": Decimal(markets_info["maintenanceMarginFraction"]),
                "clob_pair_id": markets_info["clobPairId"],
                "atomicResolution": markets_info["atomicResolution"],
                "stepBaseQuantums": markets_info["stepBaseQuantums"],
                "quantumConversionExponent": markets_info["quantumConversionExponent"],
                "subticksPerTick": markets_info["subticksPerTick"],
            }

            max_leverage = int(Decimal("1") / self._margin_fractions[trading_pair]["initial"])
            if leverage > max_leverage:
                self._perpetual_trading.set_leverage(trading_pair=trading_pair, leverage=max_leverage)
                self.logger().warning(f"Exceeded max leverage allowed."
                                      f" Leverage for {trading_pair} has been reduced to {max_leverage}")
            else:
                self._perpetual_trading.set_leverage(trading_pair=trading_pair, leverage=leverage)
                self.logger().info(f"Leverage for {trading_pair} successfully set to {leverage}.")
        return success, msg

    async def _execute_set_leverage(self, trading_pair: str, leverage: int):
        try:
            success, msg = await self._set_trading_pair_leverage(trading_pair, leverage)
            if not success:
                self.logger().network(f"Error setting leverage {leverage} for {trading_pair}: {msg}")
        except Exception:
            self.logger().network(f"Error setting leverage {leverage} for {trading_pair}")

    async def _fetch_last_fee_payment(self, trading_pair: str) -> Tuple[int, Decimal, Decimal]:
        pass

    async def _update_funding_payment(self, trading_pair: str, fire_event_on_new: bool) -> bool:
        return True



#!/usr/bin/env python
import json
from typing import TYPE_CHECKING, Dict, List, Optional

import pandas as pd

from hummingbot.client.command.gateway_api_manager import begin_placeholder_mode
from hummingbot.core.gateway.gateway_http_client import GatewayStatus
from hummingbot.core.utils.async_utils import safe_ensure_future

if TYPE_CHECKING:
    from hummingbot.client.hummingbot_application import HummingbotApplication  # noqa: F401


def ensure_gateway_online(func):
    def wrapper(self, *args, **kwargs):
        if self.trading_core.gateway_monitor.gateway_status is GatewayStatus.OFFLINE:
            self.logger().error("Gateway is offline")
            return
        return func(self, *args, **kwargs)
    return wrapper


class GatewayTokenCommand:
    """Commands for managing gateway tokens."""

    @ensure_gateway_online
    def gateway_token(self, symbol_or_address: Optional[str], action: Optional[str]):
        """
        View or update token information.
        Usage:
            gateway token <symbol_or_address>       - View token information
            gateway token <symbol> update           - Update token information
        """
        if not symbol_or_address:
            # Show help when no arguments provided
            self.notify("\nGateway Token Commands:")
            self.notify("  gateway token <symbol_or_address>       - View token information")
            self.notify("  gateway token <symbol> update           - Update token information")
            self.notify("\nExamples:")
            self.notify("  gateway token SOL")
            self.notify("  gateway token 0x1234...5678")
            self.notify("  gateway token USDC update")
            return

        if action == "update":
            safe_ensure_future(
                self._update_token_interactive(symbol_or_address),
                loop=self.ev_loop
            )
        else:
            safe_ensure_future(
                self._view_token(symbol_or_address),
                loop=self.ev_loop
            )

    async def _view_token(
        self,  # type: HummingbotApplication
        symbol_or_address: str
    ):
        """View token information across all chains."""
        try:
            # Get all available chains from the Chain enum
            from hummingbot.connector.gateway.common_types import Chain
            chains_to_check = [chain.chain for chain in Chain]
            found_tokens: List[Dict] = []

            self.notify(f"\nSearching for token '{symbol_or_address}' across all chains' default networks...")

            for chain in chains_to_check:
                # Get default network for this chain
                default_network = await self._get_gateway_instance().get_default_network_for_chain(chain)
                if not default_network:
                    continue

                # Try to get the token
                response = await self._get_gateway_instance().get_token(
                    symbol_or_address=symbol_or_address,
                    chain=chain,
                    network=default_network,
                    fail_silently=True  # Don't raise error if token not found
                )

                if "error" not in response:
                    # Extract token data - it might be nested under 'token' key
                    token_data = response.get("token", response)

                    # Add chain and network info to token data
                    token_info = {
                        "chain": chain,
                        "network": default_network,
                        "symbol": token_data.get("symbol", "N/A"),
                        "name": token_data.get("name", "N/A"),
                        "address": token_data.get("address", "N/A"),
                        "decimals": token_data.get("decimals", "N/A")
                    }
                    found_tokens.append(token_info)

            if found_tokens:
                self._display_tokens_table(found_tokens)
            else:
                self.notify(f"\nToken '{symbol_or_address}' not found on any chain's default network.")
                self.notify("You may need to add it using 'gateway token <symbol> update'")

        except Exception as e:
            self.notify(f"Error fetching token information: {str(e)}")

    async def _update_token_interactive(
        self,  # type: HummingbotApplication
        symbol: str
    ):
        """Interactive flow to update or add a token."""
        try:
            with begin_placeholder_mode(self):
                # Ask for chain
                chain = await self.app.prompt(
                    prompt="Enter chain (e.g., ethereum, solana): "
                )

                if self.app.to_stop_config or not chain:
                    self.notify("Token update cancelled")
                    return

                # Get default network for the chain
                default_network = await self._get_gateway_instance().get_default_network_for_chain(chain)
                if not default_network:
                    self.notify(f"Could not determine default network for chain '{chain}'")
                    return

                # Check if token exists
                existing_token = await self._get_gateway_instance().get_token(
                    symbol_or_address=symbol,
                    chain=chain,
                    network=default_network,
                    fail_silently=True  # Don't raise error if token not found
                )

                if "error" not in existing_token:
                    # Token exists, show current info
                    self.notify("\nCurrent token information:")
                    # Extract token data - it might be nested under 'token' key
                    token_data = existing_token.get("token", existing_token)
                    self._display_single_token(token_data, chain, default_network)

                    # Ask if they want to update
                    response = await self.app.prompt(
                        prompt="Do you want to update this token? (Yes/No) >>> "
                    )

                    if response.lower() not in ["y", "yes"]:
                        self.notify("Token update cancelled")
                        return
                else:
                    self.notify(f"\nToken '{symbol}' not found. Let's add it to {chain} ({default_network}).")

                # Collect token information
                self.notify("\nEnter token information:")

                # Symbol (pre-filled)
                token_symbol = await self.app.prompt(
                    prompt=f"Symbol [{symbol}]: "
                )
                if not token_symbol:
                    token_symbol = symbol

                # Name
                token_name = await self.app.prompt(
                    prompt="Name: "
                )
                if self.app.to_stop_config or not token_name:
                    self.notify("Token update cancelled")
                    return

                # Address
                token_address = await self.app.prompt(
                    prompt="Contract address: "
                )
                if self.app.to_stop_config or not token_address:
                    self.notify("Token update cancelled")
                    return

                # Decimals
                decimals_str = await self.app.prompt(
                    prompt="Decimals [18]: "
                )
                try:
                    decimals = int(decimals_str) if decimals_str else 18
                except ValueError:
                    self.notify("Invalid decimals value. Using default: 18")
                    decimals = 18

                # Create token data
                token_data = {
                    "symbol": token_symbol.upper(),
                    "name": token_name,
                    "address": token_address,
                    "decimals": decimals
                }

                # Display summary
                self.notify("\nToken to add/update:")
                self.notify(json.dumps(token_data, indent=2))

                # Confirm
                confirm = await self.app.prompt(
                    prompt="Add/update this token? (Yes/No) >>> "
                )

                if confirm.lower() not in ["y", "yes"]:
                    self.notify("Token update cancelled")
                    return

                # Add/update token
                self.notify("\nAdding/updating token...")
                result = await self._get_gateway_instance().add_token(
                    chain=chain,
                    network=default_network,
                    token_data=token_data
                )

                if "error" in result:
                    self.notify(f"Error: {result['error']}")
                else:
                    self.notify("✓ Token successfully added/updated!")

                    # Restart gateway for changes to take effect
                    self.notify("\nRestarting Gateway for changes to take effect...")
                    try:
                        await self._get_gateway_instance().post_restart()
                        self.notify("✓ Gateway restarted successfully")
                        self.notify(f"\nYou can now use 'gateway token {token_symbol}' to view the token information.")
                    except Exception as e:
                        self.notify(f"⚠️  Failed to restart Gateway: {str(e)}")
                        self.notify("You may need to restart Gateway manually for changes to take effect")

        except Exception as e:
            self.notify(f"Error updating token: {str(e)}")

    def _display_tokens_table(self, tokens: List[Dict]):
        """Display tokens in a table format."""
        self.notify("\nFound tokens:")

        # Create DataFrame for display
        df = pd.DataFrame(tokens)

        # Reorder columns for better display
        columns_order = ["chain", "network", "symbol", "name", "address", "decimals"]
        df = df[columns_order]

        # Format the dataframe for display
        lines = ["    " + line for line in df.to_string(index=False).split("\n")]
        self.notify("\n".join(lines))

    def _display_single_token(
        self,
        token_info: dict,
        chain: str,
        network: str
    ):
        """Display a single token's information."""
        self.notify(f"\nChain: {chain}")
        self.notify(f"Network: {network}")
        self.notify(f"Symbol: {token_info.get('symbol', 'N/A')}")
        self.notify(f"Name: {token_info.get('name', 'N/A')}")
        self.notify(f"Address: {token_info.get('address', 'N/A')}")
        self.notify(f"Decimals: {token_info.get('decimals', 'N/A')}")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: PrivateOrdersV3Api.proto
# Protobuf Python Version: 5.29.3
"""Generated protocol buffer code."""
from google.protobuf import (
    descriptor as _descriptor,
    descriptor_pool as _descriptor_pool,
    runtime_version as _runtime_version,
    symbol_database as _symbol_database,
)
from google.protobuf.internal import builder as _builder

_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    3,
    '',
    'PrivateOrdersV3Api.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18PrivateOrdersV3Api.proto\"\xe8\x05\n\x12PrivateOrdersV3Api\x12\n\n\x02id\x18\x01 \x01(\t\x12\x10\n\x08\x63lientId\x18\x02 \x01(\t\x12\r\n\x05price\x18\x03 \x01(\t\x12\x10\n\x08quantity\x18\x04 \x01(\t\x12\x0e\n\x06\x61mount\x18\x05 \x01(\t\x12\x10\n\x08\x61vgPrice\x18\x06 \x01(\t\x12\x11\n\torderType\x18\x07 \x01(\x05\x12\x11\n\ttradeType\x18\x08 \x01(\x05\x12\x0f\n\x07isMaker\x18\t \x01(\x08\x12\x14\n\x0cremainAmount\x18\n \x01(\t\x12\x16\n\x0eremainQuantity\x18\x0b \x01(\t\x12\x1d\n\x10lastDealQuantity\x18\x0c \x01(\tH\x00\x88\x01\x01\x12\x1a\n\x12\x63umulativeQuantity\x18\r \x01(\t\x12\x18\n\x10\x63umulativeAmount\x18\x0e \x01(\t\x12\x0e\n\x06status\x18\x0f \x01(\x05\x12\x12\n\ncreateTime\x18\x10 \x01(\x03\x12\x13\n\x06market\x18\x11 \x01(\tH\x01\x88\x01\x01\x12\x18\n\x0btriggerType\x18\x12 \x01(\x05H\x02\x88\x01\x01\x12\x19\n\x0ctriggerPrice\x18\x13 \x01(\tH\x03\x88\x01\x01\x12\x12\n\x05state\x18\x14 \x01(\x05H\x04\x88\x01\x01\x12\x12\n\x05ocoId\x18\x15 \x01(\tH\x05\x88\x01\x01\x12\x18\n\x0brouteFactor\x18\x16 \x01(\tH\x06\x88\x01\x01\x12\x15\n\x08symbolId\x18\x17 \x01(\tH\x07\x88\x01\x01\x12\x15\n\x08marketId\x18\x18 \x01(\tH\x08\x88\x01\x01\x12\x1d\n\x10marketCurrencyId\x18\x19 \x01(\tH\t\x88\x01\x01\x12\x17\n\ncurrencyId\x18\x1a \x01(\tH\n\x88\x01\x01\x42\x13\n\x11_lastDealQuantityB\t\n\x07_marketB\x0e\n\x0c_triggerTypeB\x0f\n\r_triggerPriceB\x08\n\x06_stateB\x08\n\x06_ocoIdB\x0e\n\x0c_routeFactorB\x0b\n\t_symbolIdB\x0b\n\t_marketIdB\x13\n\x11_marketCurrencyIdB\r\n\x0b_currencyIdB;\n\x1c\x63om.mxc.push.common.protobufB\x17PrivateOrdersV3ApiProtoH\x01P\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'PrivateOrdersV3Api_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    _globals['DESCRIPTOR']._loaded_options = None
    _globals['DESCRIPTOR']._serialized_options = b'\n\034com.mxc.push.common.protobufB\027PrivateOrdersV3ApiProtoH\001P\001'
    _globals['_PRIVATEORDERSV3API']._serialized_start = 29
    _globals['_PRIVATEORDERSV3API']._serialized_end = 773
# @@protoc_insertion_point(module_scope)



from decimal import Decimal
from typing import List, Optional, Tuple

from hummingbot.connector.exchange.paper_trade import create_paper_trade_market
from hummingbot.connector.exchange_base import ExchangeBase
from hummingbot.strategy.api_asset_price_delegate import APIAssetPriceDelegate
from hummingbot.strategy.market_trading_pair_tuple import MarketTradingPairTuple
from hummingbot.strategy.order_book_asset_price_delegate import OrderBookAssetPriceDelegate
from hummingbot.strategy.pure_market_making import InventoryCostPriceDelegate, PureMarketMakingStrategy
from hummingbot.strategy.pure_market_making.moving_price_band import MovingPriceBand
from hummingbot.strategy.pure_market_making.pure_market_making_config_map import pure_market_making_config_map as c_map


async def start(self):
    def convert_decimal_string_to_list(string: Optional[str], divisor: Decimal = Decimal("1")) -> List[Decimal]:
        '''convert order level spread string into a list of decimal divided by divisor '''
        if string is None:
            return []
        string_list = list(string.split(","))
        return [Decimal(v) / divisor for v in string_list]

    try:
        order_amount = c_map.get("order_amount").value
        order_refresh_time = c_map.get("order_refresh_time").value
        max_order_age = c_map.get("max_order_age").value
        bid_spread = c_map.get("bid_spread").value / Decimal('100')
        ask_spread = c_map.get("ask_spread").value / Decimal('100')
        minimum_spread = c_map.get("minimum_spread").value / Decimal('100')
        price_ceiling = c_map.get("price_ceiling").value
        price_floor = c_map.get("price_floor").value
        ping_pong_enabled = c_map.get("ping_pong_enabled").value
        order_levels = c_map.get("order_levels").value
        order_level_amount = c_map.get("order_level_amount").value
        order_level_spread = c_map.get("order_level_spread").value / Decimal('100')
        exchange = c_map.get("exchange").value.lower()
        raw_trading_pair = c_map.get("market").value
        inventory_skew_enabled = c_map.get("inventory_skew_enabled").value
        inventory_target_base_pct = 0 if c_map.get("inventory_target_base_pct").value is None else \
            c_map.get("inventory_target_base_pct").value / Decimal('100')
        inventory_range_multiplier = c_map.get("inventory_range_multiplier").value
        filled_order_delay = c_map.get("filled_order_delay").value
        hanging_orders_enabled = c_map.get("hanging_orders_enabled").value
        hanging_orders_cancel_pct = c_map.get("hanging_orders_cancel_pct").value / Decimal('100')
        order_optimization_enabled = c_map.get("order_optimization_enabled").value
        ask_order_optimization_depth = c_map.get("ask_order_optimization_depth").value
        bid_order_optimization_depth = c_map.get("bid_order_optimization_depth").value
        add_transaction_costs_to_orders = c_map.get("add_transaction_costs").value
        price_source = c_map.get("price_source").value
        price_type = c_map.get("price_type").value
        price_source_exchange = c_map.get("price_source_exchange").value
        price_source_market = c_map.get("price_source_market").value
        price_source_custom_api = c_map.get("price_source_custom_api").value
        custom_api_update_interval = c_map.get("custom_api_update_interval").value
        order_refresh_tolerance_pct = c_map.get("order_refresh_tolerance_pct").value / Decimal('100')
        order_override = c_map.get("order_override").value
        split_order_levels_enabled = c_map.get("split_order_levels_enabled").value
        moving_price_band = MovingPriceBand(
            enabled=c_map.get("moving_price_band_enabled").value,
            price_floor_pct=c_map.get("price_floor_pct").value,
            price_ceiling_pct=c_map.get("price_ceiling_pct").value,
            price_band_refresh_time=c_map.get("price_band_refresh_time").value
        )
        bid_order_level_spreads = convert_decimal_string_to_list(
            c_map.get("bid_order_level_spreads").value)
        ask_order_level_spreads = convert_decimal_string_to_list(
            c_map.get("ask_order_level_spreads").value)
        bid_order_level_amounts = convert_decimal_string_to_list(
            c_map.get("bid_order_level_amounts").value)
        ask_order_level_amounts = convert_decimal_string_to_list(
            c_map.get("ask_order_level_amounts").value)
        if split_order_levels_enabled:
            buy_list = [['buy', spread, amount] for spread, amount in zip(bid_order_level_spreads, bid_order_level_amounts)]
            sell_list = [['sell', spread, amount] for spread, amount in zip(ask_order_level_spreads, ask_order_level_amounts)]
            both_list = buy_list + sell_list
            order_override = {
                f'split_level_{i}': order for i, order in enumerate(both_list)
            }
        trading_pair: str = raw_trading_pair
        maker_assets: Tuple[str, str] = trading_pair.split("-")
        market_names: List[Tuple[str, List[str]]] = [(exchange, [trading_pair])]
        await self.initialize_markets(market_names)
        maker_data = [self.connector_manager.connectors[exchange], trading_pair] + list(maker_assets)
        self.market_trading_pair_tuples = [MarketTradingPairTuple(*maker_data)]

        asset_price_delegate = None
        if price_source == "external_market":
            asset_trading_pair: str = price_source_market
            ext_market = create_paper_trade_market(price_source_exchange, [asset_trading_pair])
            self.connector_manager.connectors[price_source_exchange]: ExchangeBase = ext_market
            asset_price_delegate = OrderBookAssetPriceDelegate(ext_market, asset_trading_pair)
        elif price_source == "custom_api":
            asset_price_delegate = APIAssetPriceDelegate(self.markets[exchange], price_source_custom_api,
                                                         custom_api_update_interval)
        inventory_cost_price_delegate = None
        if price_type == "inventory_cost":
            db = self.trade_fill_db
            inventory_cost_price_delegate = InventoryCostPriceDelegate(db, trading_pair)
        take_if_crossed = c_map.get("take_if_crossed").value

        should_wait_order_cancel_confirmation = c_map.get("should_wait_order_cancel_confirmation")

        strategy_logging_options = PureMarketMakingStrategy.OPTION_LOG_ALL
        self.strategy = PureMarketMakingStrategy()
        self.strategy.init_params(
            market_info=MarketTradingPairTuple(*maker_data),
            bid_spread=bid_spread,
            ask_spread=ask_spread,
            order_levels=order_levels,
            order_amount=order_amount,
            order_level_spread=order_level_spread,
            order_level_amount=order_level_amount,
            inventory_skew_enabled=inventory_skew_enabled,
            inventory_target_base_pct=inventory_target_base_pct,
            inventory_range_multiplier=inventory_range_multiplier,
            filled_order_delay=filled_order_delay,
            hanging_orders_enabled=hanging_orders_enabled,
            order_refresh_time=order_refresh_time,
            max_order_age=max_order_age,
            order_optimization_enabled=order_optimization_enabled,
            ask_order_optimization_depth=ask_order_optimization_depth,
            bid_order_optimization_depth=bid_order_optimization_depth,
            add_transaction_costs_to_orders=add_transaction_costs_to_orders,
            logging_options=strategy_logging_options,
            asset_price_delegate=asset_price_delegate,
            inventory_cost_price_delegate=inventory_cost_price_delegate,
            price_type=price_type,
            take_if_crossed=take_if_crossed,
            price_ceiling=price_ceiling,
            price_floor=price_floor,
            ping_pong_enabled=ping_pong_enabled,
            hanging_orders_cancel_pct=hanging_orders_cancel_pct,
            order_refresh_tolerance_pct=order_refresh_tolerance_pct,
            minimum_spread=minimum_spread,
            hb_app_notification=True,
            order_override={} if order_override is None else order_override,
            split_order_levels_enabled=split_order_levels_enabled,
            bid_order_level_spreads=bid_order_level_spreads,
            ask_order_level_spreads=ask_order_level_spreads,
            should_wait_order_cancel_confirmation=should_wait_order_cancel_confirmation,
            moving_price_band=moving_price_band
        )
    except Exception as e:
        self.notify(str(e))
        self.logger().error("Unknown error during initialization.", exc_info=True)
